
from twisted.internet import gtk2reactor
gtk2reactor.install()
from twisted.internet import reactor, defer
from twisted.python import log

from petmail import agent
from petmail import mm_util
from petmail.gui.ticketAgent import TicketAgent

import pygtk
pygtk.require("2.0")
import gtk, gtk.glade, gobject

import pickle, os, os.path, sys, time
#log.startLogging(sys.stderr)

def sibling(name):
    script = __file__
    d = os.path.dirname(script)
    return os.path.join(d, name)

gladefile = sibling("gui.glade")

def set_image(widget, data):
    if data:
        try:
            loader = gtk.gdk.gdk_pixbuf_loader_new()
            # apparently gdk_pixbuf_loader_set_size isn't bound yet
            #loader.set_size(64, 64)
            loader.write(data)
            loader.close()
            pb = loader.get_animation()
        except gobject.GError, e:
            print "unloadable image data"
            print e
        else:
            if pb.is_static_image():
                static = pb.get_static_image()
                widget.set_from_pixbuf(static)
            else:
                widget.set_from_animation(pb)
            widget.show()
    else:
        widget.set_from_file("") # to free the memory / animation CPU cycles
        widget.hide()
    
class Helper(object):

    def set_sensitive(self, names, sensitive):
        for name in names:
            try:
                self.w(name).set_sensitive(sensitive)
            except AttributeError, e:
                e.args += ("[%s]" % name,)
                raise e

    def w(self, name):
        return self.xml.get_widget(name)
    def text(self, name):
        return self.w(name).get_text()

class LogWindow(Helper):
    def __init__(self, gui):
        xml = gtk.glade.XML(gladefile, "log")
        self.xml = xml
        self.parent = gui
        self.gui = gui
        self.model = gtk.TreeStore(str, str, object)
        xml.signal_connect("do_log_close", self.do_close)
        xml.signal_connect("do_log_clear", self.do_clear)
        xml.signal_connect("do_log_append", self.do_append)

        view = self.w("log_view")
        view.set_model(self.model)
        self.view = view

        col = gtk.TreeViewColumn("Time", gtk.CellRendererText(), text=0)
        col.set_resizable(1)
        view.append_column(col)

        col = gtk.TreeViewColumn("Event", gtk.CellRendererText(), text=1)
        col.set_resizable(1)
        view.append_column(col)

        view.get_selection().set_mode(gtk.SELECTION_NONE)
        view.connect("button_press_event", self.do_clicked)

        self.window = self.w("log")
        self.window.connect("destroy", self.do_destroy)
        self.fill()

    def stringify(self, message):
        return ' '.join(map(str, message))

    def render(self, event):
        timestamp = event.time
        #date = time.strftime("%d %b %Y", time.localtime(timestamp))
        t = time.strftime("%H:%M:%S", time.localtime(timestamp))
        text = self.stringify(event.getMessage())
        return (t, text, event)
        
    def fill(self):
        self.model.clear()
        l = self.gui.agent.eventLog
        for event in l.getItems(None):
            self.fill_entry(None, event)
        l.addObserver(self.add)
        self.scroll_to_bottom()

    def fill_entry(self, viewparent, lognode):
        fields = self.render(lognode)
        row = self.model.append(viewparent, fields)
        for child in lognode.getChildren():
            self.fill_entry(row, child)


    def add_helper(self, model, path, iter, data):
        parent, fields, found = data
        if model[path][2] == parent:
            model.append(iter, fields)
            # do we need to emit a signal here? apparently not.
            found[0] = True
            return True
        return False

    def add(self, event, pclass, parent):
        fields = self.render(event)
        if parent:
            # find the parent. It might be in a sub-node, so we must search
            # recursively. Ick. This will eventually be replaced with
            # something more efficient.
            found = [False]
            self.model.foreach(self.add_helper, (parent, fields, found))
            if not found[0]:
                print "HEY!, no parent", fields
                print " parent=", self.stringify(parent.getMessage())
            return
        self.model.append(None, fields)
        self.scroll_to_bottom()

    def scroll_to_bottom(self):
        row = self.model[len(self.model)-1]
        self.view.scroll_to_cell(row.path)

    def show(self):
        self.window.deiconify()
    def do_clicked(self, *args):
        #print "clicked", args
        pass

    def do_clear(self, widget):
        self.model.clear()

    def do_append(self, widget):
        self.gui.agent.log("adding fake line")

    def do_close(self, widget):
        self.window.destroy()
    def do_destroy(self, widget):
        self.gui.agent.eventLog.removeObserver(self.add)
        self.gui.event_log_closed()

class InletPanel(Helper):
    settings_panel = None

    def __init__(self, mepane, oldinlet=None):
        self.parent = mepane
        self.xml = gtk.glade.XML(gladefile, "inlet_config")
        self.window = self.w("inlet_config")
        self.combo = self.w("inlet_config_type")
        self.table = self.w("inlet_config_table")
        self.xml.signal_connect("do_inlet_newtype", self.do_type)
        self.xml.signal_connect("do_inlet_config_cancel", self.do_cancel)
        self.xml.signal_connect("do_inlet_config_ok", self.do_ok)
        self.combo.set_property("active", 0)

        if oldinlet:
            self.populate(oldinlet)

    def populate(self, oldinlet):
        itype = oldinlet.getType()
        print "populate(%s)" % itype
        self.itype = itype
        combo_labels = {'dummy': 0,
                        'pb': 1,
                        'maildir': 2,
                        'mailsocket': 3,
                        'imap': 4,
                        'pop': 5,
                        'jabber': 6}
        which = combo_labels[itype]
        self.combo.set_property("active", which)
        #self.load_settings_panel(itype)

        w = self.settings_panel_xml.get_widget

        if itype == "jabber":
            fields = {
                'owner_inlet_jabber_jid_hostname': "server",
                'owner_inlet_jabber_jid_username': "username",
                'owner_inlet_jabber_password': "passwd",
                'owner_inlet_jabber_host': "host",
                'owner_inlet_jabber_port': "port",
                }
        elif itype == "maildir":
            fields = {
                'owner_inlet_address': 'address',
                'owner_inlet_maildir_directory': 'basdir',
                }
        elif itype == "mailsocket":
            fields = {
                'owner_inlet_address': 'address',
                'owner_inlet_mailsocket_port': 'portnum',
                }
        elif itype in ("pop3", "imap"):
            fields = {
                'owner_inlet_address': 'address',
                'owner_inlet_host': 'host',
                'owner_inlet_port': 'portnum',
                'owner_inlet_username': 'username',
                'owner_inlet_password': 'password',
                'owner_inlet_pollinterval': 'interval',
                }
            if itype == "imap":
                fields['owner_inlet_folder'] = 'folder'
        elif itype == 'pb':
            fields = {
                'owner_inlet_pb_hostname': 'hostname',
                'owner_inlet_pb_port': 'portnum',
                }
        else:
            log.msg("unknown itype '%s'" % itype)
            return

        for k,v in fields.items():
            w(k).set_text(str(oldinlet.getProperty(v)))


    def start(self):
        self.d = defer.Deferred()
        return self.d

    def load_settings_panel(self, itype):
        ptype = itype + "_settings"
        if itype == "pop3":
            ptype = "imap_settings" # uses a slightly modified imap_settings
        x = gtk.glade.XML(gladefile, ptype)
        w = x.get_widget(ptype)
        assert w
        if itype == "pop3":
            x.get_widget("imap_settings_name").set_text("POP3 Settings")
            x.get_widget("owner_inlet_folder_label").set_sensitive(0)
            x.get_widget("owner_inlet_folder").set_sensitive(0)
        if self.settings_panel:
            self.table.remove(self.settings_panel)
        self.settings_panel_xml, self.settings_panel = x,w
        self.table.attach(w, 0, 1, 1, 2)

    def do_type(self, combo):
        m = combo.get_model()
        name = m[combo.get_property("active")][0].lower()
        itype = name
        if itype == "pop":
            itype = "pop3"
        self.itype = itype
        self.load_settings_panel(itype)

    def do_cancel(self, widget):
        self.window.destroy()
        self.d.callback("cancelled")

    def do_ok(self, widget):
        iargs = {}
        w = self.settings_panel_xml.get_widget
        assert self.itype in ("pop3", "imap", "mailsocket", "maildir",
                              "pb", "jabber", "dummy")
        if self.itype in ("pop3", "imap", "mailsocket", "maildir"):
            iargs['address'] = w("owner_inlet_address").get_text()
        if self.itype in ("pop3", "imap"):
            iargs['host'] = w("owner_inlet_host").get_text()
            iargs['portnum'] = int(w("owner_inlet_port").get_text())
            iargs['username'] = w("owner_inlet_username").get_text()
            iargs['password'] = w("owner_inlet_password").get_text()
            iargs['interval'] = int(w("owner_inlet_pollinterval").get_text())
            if self.itype == "imap":
                iargs['folder'] = w("owner_inlet_folder").get_text()

        if self.itype == "maildir":
            iargs['basedir'] = w("owner_inlet_maildir_directory").get_text()

        if self.itype == "mailsocket":
            port = w("owner_inlet_mailsocket_port").get_text()
            iargs['portnum'] = int(port)

        if self.itype == "pb":
            iargs['hostname'] = w("owner_inlet_pb_hostname").get_text()
            iargs['portnum'] = int(w("owner_inlet_pb_port").get_text())

        if self.itype == "jabber":
            fields = {
                'server': "owner_inlet_jabber_jid_hostname",
                'username': "owner_inlet_jabber_jid_username",
                'passwd': "owner_inlet_jabber_password",
                'host': "owner_inlet_jabber_host",
                'port': "owner_inlet_jabber_port",
                }
            for k,v in fields.items():
                iargs[k] = w(v).get_text()
            iargs['port'] = int(iargs['port'])

        inlet = None
        if self.itype != "dummy":
            inlet = self.parent.gui.agent.newInlet(self.itype, **iargs)

        self.window.destroy()
        self.d.callback(inlet)

    def do_popup_type(self, widget, event):
        print "do_popup_type", event.type
        if event.type == gtk.gdk.BUTTON_PRESS:
            #widget.popup(None, None, None, event.button, event.time)
            self.popup.popup(None, None, None, event.button, event.time)

    def do_change_type(self, itype):
        print "do_type2", itype
        # populate the second field with a suitable config widget
        if itype == "jabber":
            w = gtk.glade.XML(gladefile, "jabber_settings")
            self.w("inlet_config_table").attach(w, 0, 0, 1, 1)

class OwnerWindow(Helper):
    def __init__(self, mepane, owner, new):
        self.mepane = mepane
        self.gui = mepane.gui
        self.owner = owner
        self.new = new

        xml = gtk.glade.XML(gladefile, "owner")
        self.xml = xml

        g = xml.get_widget
        self.window = g("owner")
        self.delete_button = g("my_delete_button")
        self.ok_button = g("my_ok_button")

        if new:
            self.ok_button.set_label("Add")
            self.delete_button.set_sensitive(False)
        else:
            self.ok_button.set_label("Update")
            self.ok_button.set_sensitive(False)
            self.delete_button.set_sensitive(True)

        inletview = g("inlets")
        self.inlet_list = gtk.ListStore(str, object)
        inletview.set_model(self.inlet_list)
        inletview.append_column(gtk.TreeViewColumn("Inlet",
                                                   gtk.CellRendererText(),
                                                   text=0))

        server_fields = ["owner_inlet_host_label", "owner_inlet_host",
                         "owner_inlet_port_label", "owner_inlet_port",
                         "owner_inlet_username_label",
                         "owner_inlet_username",
                         "owner_inlet_password_label",
                         "owner_inlet_password",
                         "owner_inlet_pollinterval_label",
                         "owner_inlet_pollinterval",
                         "owner_inlet_pollnow"]
        self.server_fields = server_fields

        email_toggles = ["owner_inlet_imap", "owner_inlet_pop3",
                         "owner_inlet_maildir", "owner_inlet_mailsocket"]
        self.email_toggles = email_toggles

        c = xml.signal_connect
        c("do_dirty", self.do_dirty)
        c("on_owner_is_secure_toggled", self.secure_toggled)

        c("do_me_delete_inlet", self.inlet_delete)
        c("do_me_modify_inlet", self.inlet_modify)
        c("do_me_add_inlet", self.inlet_add)

        c("on_owner_inlet_email_toggled", self.inlet_email_toggled)
        c( "on_owner_inlet_imap_toggled", self.inlet_imap_toggled)
        c( "on_owner_inlet_pop3_toggled", self.inlet_pop3_toggled)
        c("on_owner_inlet_pollnow_clicked", self.inlet_pollnow)
        c( "on_owner_inlet_maildir_toggled", self.inlet_maildir_toggled)
        c( "on_owner_inlet_mailsocket_toggled",
           self.inlet_mailsocket_toggled)
        c("on_owner_inlet_pb_toggled", self.inlet_pb_toggled)
        c("on_owner_inlet_jabber_toggled", self.inlet_jabber_toggled)
        c("on_owner_inlet_none_toggled", self.inlet_none_toggled)
        c("on_my_maxsize_toggle_toggled", self.maxsize_toggled)

        c("on_owner_image_choose_clicked", self.do_image)
        c("on_owner_image_remove_clicked", self.do_image_remove)

        c("do_me_cancel", self.do_close)
        c("do_me_armored", self.do_armor)
        c("do_me_delete", self.do_delete)
        c("do_me_ok", self.do_save)

        self.window.connect("destroy", self.do_destroy)
        self.image_filename = None
        self.populate()


    def populate(self):
        owner = self.owner
        self.window.set_title("My Identity '%s'" % owner.getName())

        self.w("my_ID_label").set_text(owner.ID)

        myname = owner.getProperty("myName") or ""
        self.w("my_local_name_entry").set_text(myname)

        self.w("owner_is_secure").set_active(owner.getProperty("pseudonymous"))

        names = owner.getProperty("name") or ""
        self.w("my_name_entry").set_text(names)

        atext = ", ".join(owner.getProperty("addresses"))
        self.w("my_addresses_entry").set_text(atext)

        self.inlet_list.clear()
        for inlet in owner.getInlets():
            self.inlet_list.append((inlet.encode(), inlet))

        if 0:
            inlets = owner.getInlets()
            assert len(inlets) <= 1 # TODO: enhance gui to handle many inlets
            if inlets:
                inlet = inlets[0]
                itype = inlet.getType()
                ip = inlet.getProperty
            else:
                itype = None
                inlet = None

            self.w("owner_inlet_email").set_active(0)
            self.w( "owner_inlet_imap").set_active(0)
            self.w( "owner_inlet_pop3").set_active(0)
            self.w( "owner_inlet_maildir").set_active(0)
            self.w( "owner_inlet_mailsocket").set_active(0)
            self.w("owner_inlet_pb").set_active(0)
            self.w("owner_inlet_jabber").set_active(0)
            self.w("owner_inlet_none").set_active(0)

            if itype in ("maildir", "mailsocket", "imap", "pop3"):
                # all of these inlets show an email address to the outside world
                self.w("owner_inlet_email").set_active(1)
                self.w("owner_inlet_address").set_text(ip("address"))

                if itype == "maildir":
                    self.w("owner_inlet_maildir").set_active(1)
                    w = self.w("owner_inlet_maildir_directory")
                    w.set_text(ip("basedir"))
                elif itype == "mailsocket":
                    self.w("owner_inlet_mailsocket").set_active(1)
                    port = str(ip("portnum"))
                    self.w("owner_inlet_mailsocket_port").set_text(port)
                else:
                    # populate server fields for imap/pop3
                    self.w("owner_inlet_host").set_text(ip("host"))
                    self.w("owner_inlet_port").set_text("%d" % ip("portnum"))
                    self.w("owner_inlet_username").set_text(ip("username"))
                    # TODO: to show the password or not?
                    self.w("owner_inlet_password").set_text(ip("password"))
                    w = self.w("owner_inlet_pollinterval")
                    w.set_value(ip("pollinterval"))
                    if itype == "imap":
                        self.w("owner_inlet_imap").set_active(1)
                        self.w("owner_inlet_folder").set_text(ip("folder"))
                    elif itype == "pop3":
                        self.w("owner_inlet_pop3").set_active(1)

            elif itype == "pb":
                self.w("owner_inlet_pb").set_active(1)
                self.w("owner_inlet_pb_hostname").set_text(ip("hostname"))
                self.w("owner_inlet_pb_port").set_text(str(ip("portnum")))
            elif itype == "jabber":
                self.w("owner_inlet_jabber").set_active(1)
                self.w("owner_inlet_jabber_jid_username").set_text(ip("username"))
                self.w("owner_inlet_jabber_jid_hostname").set_text(ip("server"))
                self.w("owner_inlet_jabber_password").set_text(ip("passwd"))
                self.w("owner_inlet_jabber_host").set_text(ip("host"))
                self.w("owner_inlet_jabber_port").set_text(str(ip("port")))
            else:
                self.w("owner_inlet_none").set_active(1)

        gp = owner.getGenericPermission("rfc822")
        assert gp

        maxsize = gp.getProperty("maxsize")
        if maxsize is not None:
            self.w("my_maxsize_toggle").set_active(1)
            maxsize = float(maxsize / 1000)
            self.w("my_maxsize_entry").set_value(maxsize)
        else:
            self.w("my_maxsize_toggle").set_active(0)

        self.w("my_accept_html").set_active(gp.getProperty("acceptHTML"))

        url_string = " ".join(gp.listTicketTakerURLs())
        self.w("my_ticket_urls_entry").set_text(url_string)

        if 0:
            for n in ["owner_inlet_email", "owner_inlet_imap",
                      "owner_inlet_pop3", "owner_inlet_maildir",
                      "owner_inlet_mailsocket", "owner_inlet_pb",
                      "owner_inlet_jabber", "owner_inlet_none",
                      "my_maxsize_toggle"]:
                self.w(n).emit("toggled")

        #set_image(self.w("owner_image"), owner.extensions.get("image"))
            
        self.ok_button.set_sensitive(0)
        if 0 and (self.w("owner_inlet_email").get_active() and
            self.w("owner_inlet_imap").get_active()):
            self.w("owner_inlet_pollnow").set_sensitive(1)

    def secure_toggled(self, widget):
        # this is done immediately, rather than at do_close(), because it
        # can raise an exception. TODO: do this better.
        self.owner.setPseudonymous(widget.get_active())
        # TODO: this may raise a CatOutOfBagError, in which case the correct
        # behavior is to present a dialog box to give the user the option of
        # setting it anyway, by giving yesReally=True to the call.

    def inlet_delete(self, widget):
        sel = self.w("inlets").get_selection()
        model, iter = sel.get_selected()
        model.remove(iter)
        self.do_dirty()

    def inlet_modify(self, widget):
        sel = self.w("inlets").get_selection()
        model, iter = sel.get_selected()
        oldinlet = model[iter][1]
        p = InletPanel(self, oldinlet)
        d = p.start()
        d.addCallback(self._modified_inlet, oldinlet)

    def _modified_inlet(self, newinlet, oldinlet):
        for row in iter(self.inlet_list):
            if row[1] is oldinlet:
                row[0] = newinlet.encode() # TODO: not in the Interface
                row[1] = newinlet
                self.do_dirty()
                return
        log.msg("weird, inlet disappeared before we could modify it")

    def inlet_add(self, widget):
        p = InletPanel(self)
        d = p.start()
        d.addCallback(self._add_inlet)

    def _add_inlet(self, newinlet):
        if newinlet == "cancelled":
            return
        self.inlet_list.append((newinlet.encode(), newinlet))
        self.do_dirty()

    def inlet_email_toggled(self, widget):
        active = widget.get_active()
        self.set_sensitive(self.email_toggles, active)
        self.set_sensitive(["owner_inlet_address_label",
                            "owner_inlet_address"], active)
        for n in self.email_toggles:
            self.w(n).emit("toggled")

    def inlet_imap_toggled(self, widget):
        server_active = (self.w("owner_inlet_email").get_active() and
                         (widget.get_active() or
                          self.w("owner_inlet_pop3").get_active()))
        self.set_sensitive(self.server_fields, server_active)
        imap_active = (self.w("owner_inlet_email").get_active() and
                       widget.get_active())
        self.set_sensitive(["owner_inlet_folder_label",
                            "owner_inlet_folder"], imap_active)
        if imap_active:
            if self.w("owner_inlet_port").get_text() == "110":
                self.w("owner_inlet_port").set_text("143")

    def inlet_pop3_toggled(self, widget):
        server_active = (self.w("owner_inlet_email").get_active() and
                         (widget.get_active() or
                          self.w("owner_inlet_imap").get_active()))
        self.set_sensitive(self.server_fields, server_active)
        pop3_active = (self.w("owner_inlet_email").get_active() and
                       widget.get_active())
        if pop3_active:
            if self.w("owner_inlet_port").get_text() == "143":
                self.w("owner_inlet_port").set_text("110")

    def inlet_maildir_toggled(self, widget):
        active = (widget.get_active() and
                  self.w("owner_inlet_email").get_active())
        fields = ["owner_inlet_maildir_directory_label",
                  "owner_inlet_maildir_directory"]
        self.set_sensitive(fields, active)

    def inlet_mailsocket_toggled(self, widget):
        active = (widget.get_active() and
                  self.w("owner_inlet_email").get_active())
        fields = ["owner_inlet_mailsocket_port_label",
                  "owner_inlet_mailsocket_port"]
        self.set_sensitive(fields, active)

    def inlet_pb_toggled(self, widget):
        fields = ["owner_inlet_pb_hostname_label",
                  "owner_inlet_pb_hostname",
                  "owner_inlet_pb_port_label",
                  "owner_inlet_pb_port"]
        self.set_sensitive(fields, widget.get_active())

    def inlet_jabber_toggled(self, widget):
        fields = ["owner_inlet_jabber_jid_username",
                  "owner_inlet_jabber_jid_hostname",
                  "owner_inlet_jabber_password",
                  "owner_inlet_jabber_host",
                  "owner_inlet_jabber_port"]
        self.set_sensitive(fields, widget.get_active())

    def inlet_none_toggled(self, widget):
        fields = ["owner_inlet_none_label"]
        self.set_sensitive(fields, widget.get_active())

    def maxsize_toggled(self, widget):
        self.set_sensitive(["my_maxsize_entry"], widget.get_active())

    def inlet_pollnow(self, widget):
        # quickie debug hack to let any exceptions get displayed
        for i in self.owner.getInlets():
            if i.getType() in ("IMAP", "POP3"):
                print "polling", i
                d = i.check()
                d.addCallback(self._poll_done, i)
                d.addErrback(self._poll_failed, i)

    def _poll_done(self, results, inlet):
        win = self.w("owner")
        m = gtk.MessageDialog(win, gtk.DIALOG_DESTROY_WITH_PARENT,
                              gtk.MESSAGE_WARNING, gtk.BUTTONS_CLOSE,
                              "%s polled" % inlet)
        m.connect("response", lambda m,y: m.destroy())
        m.show()

    def _poll_failed(self, why, inlet):
        win = self.w("owner")
        msg = "%s poll failed\n%s\n%s" % (inlet,
                                          why.getErrorMessage(),
                                          why.getBriefTraceback())
        m = gtk.MessageDialog(win, gtk.DIALOG_DESTROY_WITH_PARENT,
                              gtk.MESSAGE_WARNING, gtk.BUTTONS_CLOSE,
                              msg)
        m.connect("response", lambda m,y: m.destroy())
        m.show()

    def do_image(self, widget):
        self.fs = fs = gtk.FileSelection("Choose Image")
        fs.hide_fileop_buttons()
        fs.ok_button.connect("clicked", self.do_image_ok)
        fs.cancel_button.connect("clicked", self.do_image_cancel)
        if self.image_filename and self.image_filename != ('removed',):
            fs.set_property('filename', self.image_filename)
        fs.show()
    def do_image_ok(self, widget):
        filename = self.fs.get_property('filename')
        data = open(filename, "r").read()
        if len(data) > 640*1024:
            print "image data too large"
            self.fs.destroy()
            self.fs = None
            return
        try:
            pb = gtk.gdk.PixbufAnimation(filename)
        except gobject.GError, e:
            print "unloadable image file"
            print e
            # TODO: pop up a dialog
        else:
            x,y = pb.get_width(), pb.get_height()
            if max(x,y) > 128:
                print "please limit yourself to 128x128"
            widget = self.w("owner_image")
            if pb.is_static_image():
                widget.set_from_pixbuf(pb.get_static_image())
            else:
                self.w("owner_image").set_from_animation(pb)
            widget.show()
            self.image_filename = filename
            self.do_dirty()
        self.fs.destroy()
        self.fs = None
    def do_image_cancel(self, widget):
        self.fs.destroy()
        self.fs = None
    def do_image_remove(self, widget):
        set_image(self.w("owner_image"), None)
        self.image_filename = ("removed",) # special marker
        self.do_dirty()

    def do_close(self, widget):
        # close the window without making any changes
        self.window.destroy()

    def do_destroy(self, widget):
        # the window is being removed, we must let the MePane know
        self.mepane.closeOwner(self.owner)
        return False # do normal handling too

    def show(self):
        w = self.xml.get_widget("owner")
        w.deiconify() # this also appears to raise it to the top

    def do_dirty(self, *args):
        self.w("my_cancel_button").set_label("Cancel")
        #self.w("owner_inlet_pollnow").set_sensitive(0)
        self.ok_button.set_sensitive(1)

    def do_delete(self, widget):
        # confirm
        xml = gtk.glade.XML(gladefile, "confirm_dialog")
        dialog = xml.get_widget("confirm_dialog")
        l = xml.get_widget("confirm_label")
        confirm = ""
        confirm += "All secret keys will be deleted.\n"
        confirm += "All old mail will be deleted.\n" # TODO: really?
        confirm += "All Permissions will be deleted.\n"
        confirm += "You will no longer be able to receive any mail"
        confirm += " addressed to this Identity in the future. Ever.\n"
        confirm += "(just to be sure, we will notify all current Permission"
        confirm += " holders and ask them to delete this entry from their"
        confirm += " addressbooks)\n"
        confirm += "\n"
        confirm += "<b>Really delete Identity '%s'?</b>\n" \
                   % self.owner.getName()
        l.set_markup(confirm)
        dialog.set_title("Delete Identity %s" % self.owner.ID)
        dialog.connect("response", self._confirm_response)
    def _confirm_response(self, dialog, responseid):
        if responseid == gtk.RESPONSE_OK:
            self.really_delete()
        dialog.destroy()
    def really_delete(self):
        self.gui.agent.removeOwner(self.owner)
        self.window.destroy()
        #self.mepane.closeOwner(self.owner)
        self.mepane.ownerModified(self.owner)

    def do_save(self, widget):
        owner = self.owner

        # addresses
        text = self.w("my_addresses_entry").get_text()
        pieces = text.split(",")
        new = []
        for i in pieces:
            i = i.strip()
            if i:
                new.append(i)
        owner.setProperty("addresses", new)

        # local name
        myName = self.w("my_local_name_entry").get_text()
        owner.setProperty("myName", myName.strip())

        # published name
        text = self.w("my_name_entry").get_text()
        owner.setProperty("name", text.strip())

        # inlet
        inlets = [row[1] for row in list(self.inlet_list)]
        # TODO: this might unnecessarily interrupt an existing inlet..
        # depends upon how clever setInlets() cares to be
        owner.setInlets(inlets)

        # our GenericPermission fields
        gp = owner.genericPermission
        if self.w("my_maxsize_toggle").get_active():
            maxsize = int(self.w("my_maxsize_entry").get_value()) * 1000
            gp.setMaxSize(maxsize)
        else:
            gp.setMaxSize(None)
        if self.w("my_accept_html").get_active():
            gp.setContentTypes(["*/*"])
        else:
            gp.setContentTypes(["text/plain"]) # TODO: needs work
        urlString = self.w("my_ticket_urls_entry").get_text()
        gp.setTicketTakersFromString(self.gui.agent, urlString)

        if self.image_filename in [('removed',), None]:
            owner.extensions['image'] = None
        else:
            imagedata = open(self.image_filename, "r").read()
            if len(imagedata) > 640*1024:
                print "image data too large"
            else:
                owner.extensions['image'] = imagedata

        if self.new:
            self.gui.agent.addIdentity(owner) # will publish
            #self.ok_button.set_label("Update")
            #self.delete_button.set_sensitive(True)
            self.window.destroy()
        else:
            owner.publish() # TODO: will auto-publish take care of this??
            self.w("my_cancel_button").set_label("Close")
            self.ok_button.set_sensitive(0)

        self.gui.scheduleSave()
        self.mepane.ownerModified(self.owner)

    def do_armor(self, widget):
        d = self.owner.getArmored()
        d.addCallback(self._printArmored)

    def _printArmored(self, armor):
        print armor


class KeyGenerationProgress(object):
    def __init__(self, window, progress_widget, entropy_widget):
        self.window = window
        self.progress = progress_widget
        self.entropy = entropy_widget
        self.do_entropy = False
    def set_progress(self, progress):
        self.progress.set_fraction(progress)
    def pulse_progress(self):
        self.progress.pulse()
        # TODO: widget seems to lose the text here
    def set_entropy(self, entropy):
        if not self.do_entropy:
            self.do_entropy = True
            self.entropy.show()
        self.entropy.set_fraction(entropy)
    def done(self):
        self.window.destroy()

class MePane(object):
    def __init__(self, gui):
        self.gui = gui
        self.ownerpanels = {}

    def load(self, xml):
        g = xml.get_widget
        xml.signal_connect("do_create_identity", self.do_create_identity)
        self.app = g('app')
        self.view = g('owner_tree')
        model = self.model = gtk.ListStore(gobject.TYPE_STRING,
                                           gobject.TYPE_STRING,
                                           gobject.TYPE_PYOBJECT)
        self.view.set_model(model)

        col = gtk.TreeViewColumn("Name", gtk.CellRendererText(), text=0)
        col.set_resizable(1)
        self.view.append_column(col)

        col = gtk.TreeViewColumn("ID", gtk.CellRendererText(), text=1)
        col.set_resizable(1)
        self.view.append_column(col)

        sel = self.view.get_selection()
        sel.set_mode(gtk.SELECTION_SINGLE)
        self.view.connect("button_press_event", self.do_clicked)


    def populate(self):
        self.model.clear()
        for owner in self.gui.agent.getAllIdentities():
            i = self.model.append(None)
            self.model.set(i,
                           0, owner.getProperty("myName"),
                           1, owner.getProperty("id"),
                           2, owner)

    def do_clicked(self, widget, event):
        if event.type != gtk.gdk._2BUTTON_PRESS:
            return
        path, col, cx, cy = widget.get_path_at_pos(int(event.x),
                                                   int(event.y))
        sel = widget.get_selection()
        sel.select_path(path)
        model, iter = sel.get_selected()
        name = model.get_value(iter, 0)
        owner = model.get_value(iter, 2)
        self.inspectOwner(owner, new=False)

    def do_create_identity(self, widget):
        x = gtk.glade.XML(gladefile, "generateKey")
        p0 = x.get_widget("generateKey")
        p1 = x.get_widget("key_generation_progress")
        p2 = x.get_widget("entropy_level")
        p = KeyGenerationProgress(p0, p1, p2)
        
        d = self.gui.agent.newIdentity(p)
        d.addCallback(self._create1)
        d.addErrback(self._create_failed)

    def _create1(self, owner):
        # add permission
        p = self.gui.agent.newPermission("rfc822")
        owner.setGenericPermission("rfc822", p)
        self.inspectOwner(owner, new=True)

    def _create_failed(self, why):
        m = gtk.MessageDialog(self.app, gtk.DIALOG_DESTROY_WITH_PARENT,
                              gtk.MESSAGE_WARNING, gtk.BUTTONS_CLOSE,
                              "Unable to create Identity: %s" % why)
        m.connect("response", lambda m,y: m.destroy())
        m.show()

    def inspectOwner(self, owner, new):
        w = self.ownerpanels.get(owner)
        if not w:
            w = OwnerWindow(self, owner, new)
            self.ownerpanels[owner] = w
            w.populate()
        w.show() # makes sure it is on top, and not iconified

    def closeOwner(self, owner):
        # this is called after the OwnerWindow has been closed
        del self.ownerpanels[owner]

    def ownerModified(self, owner):
        # note! owner may no longer exist (we can be called in response to a
        # delete action)
        self.populate()
        # TODO: update selection
        self.gui.addrpane.populateOwners()
        self.gui.sendpane.populateOwners()
        self.gui.inboxpane.populateOwners()

class AddressBookPane(Helper):
    frozen = 0
    needUpdate = False
    currentEntry = None
    owner = None

    def __init__(self, gui):
        self.gui = gui

    def load(self, xml):
        self.xml = xml
        # addressbook_tree, ID_entry, names_entry, petname_entry, IMT_entry,
        # maxsize_entry, accept_html
        # outbound_IMT, outbound_maxsize, outbound_html
        view = self.addressbook_tree = xml.get_widget("addressbook_tree")
        model = self.addressbook_list = gtk.ListStore(gobject.TYPE_STRING,
                                                      gobject.TYPE_PYOBJECT)
        view.set_model(model)
        view.append_column(gtk.TreeViewColumn("Name",
                                              gtk.CellRendererText(),
                                              text=0))
        sel = view.get_selection()
        sel.set_mode(gtk.SELECTION_SINGLE)
        sel.connect("changed", self.do_select)
        #self.populateAddressBook()
        g = xml.get_widget
        c = xml.signal_connect

        self.w("petname_entry").connect("activate",
                                        self.petname_entry_changed)

        c("do_lookup", self.do_lookup)
        c("on_maxsize_toggle_toggled", self.do_maxsize_toggled)
        c("on_IMT_toggle_toggled", self.do_IMG_toggled)
        c("do_entry_dirty", self.entry_dirty)
        c("on_addressbook_entry_cancel_clicked", self.do_cancel)
        c("on_addressbook_entry_delete_clicked", self.do_delete)
        c("on_addressbook_entry_ping_clicked", self.do_ping)
        c("on_addressbook_entry_send_clicked", self.do_send)
        c("on_addressbook_entry_ok_clicked", self.do_ok)
        c("on_grant_specific_toggled", self.do_specific_toggled)

        self.populateOwners()
        c("on_addressbook_owner_entry_changed", self.do_owner)

        owners = self.gui.agent.getAllIdentities()
        if owners:
            self.setOwner(owners[0])
        self.do_select(sel)

    def populateOwners(self):
        owners = self.gui.getIdentityNames()
        if not owners:
            owners = [""]
        default = owners[0]
        self.w("addressbook_owner").set_popdown_strings(owners)
        self.w("addressbook_owner_entry").set_text(default)

    def do_owner(self, widget):
        # This is hooked to the GtkEntry's "changed" signal. When the entry
        # is changed by arrow-key-stepping through the GtkCombo, the
        # following signals are emitted:
        #  delete_text, changed, insert_text, changed
        # So we will be called twice, once while the Entry is empty.
        ownername = widget.get_text()
        if ownername != "":
            owner = self.gui.getOwnerByName(ownername)
            self.setOwner(owner)

    def jumpToOwner(self, owner):
        self.w("addressbook_owner_entry").set_text(owner.getName())
        # it appears that this generates the right signal

    def setOwner(self, owner):
        if owner != self.owner and self.owner:
            oldbook = self.owner.getAddressBook()
            oldbook.unsubscribe(self.updated)
        self.owner = owner

        book = self.owner.getAddressBook()
        book.subscribe(self.updated)

        self.populate()

    def do_lookup(self, widget):
        addr = self.w("lookup_entry").get_text()
        if addr:
            book = self.owner.getAddressBook()
            d = book.addAddress(addr)
            d.addCallback(self._lookup_complete1)
            d.addErrback(log.err)
    def _lookup_complete1(self, entry):
        self.w("lookup_entry").set_text("")
        # try to select that entry, will do a self.populate()
        self.go_to_entry(entry)

    def populate(self):
        book = self.owner.getAddressBook()
        self.addressbook_list.clear()
        for e in book.entries:
            iter = self.addressbook_list.append()
            self.addressbook_list.set(iter,
                                      0, e.getName(),
                                      1, e)
        #self.addressbook_tree.get_selection().select_path([0])
        self.addressbook_tree.get_selection().unselect_all()

    def go_to_entry(self, entry):
        """This selects both the owner and the entry."""
        owner = entry.getOwner()
        self.jumpToOwner(owner) # this populates the list
        model = self.addressbook_list
        for row in model:
            if row[1] == entry:
                self.addressbook_tree.scroll_to_cell(row.path)
                self.addressbook_tree.get_selection().select_path(row.path)
                return

    def do_specific_toggled(self, widget):
        specific_active = widget.get_active()
        disabled = ["IMT_toggle", "maxsize_toggle", "accept_html",
                    "grant_3rd_party"]
        self.set_sensitive(disabled, specific_active)
        IMT_active = self.w("IMT_toggle").get_active()
        if not specific_active:
            IMT_active = 0
        self.w("IMT_entry").set_sensitive(IMT_active)
        maxsize_active = self.w("maxsize_toggle").get_active()
        if not specific_active:
            maxsize_active = 0
        self.w("maxsize_entry").set_sensitive(maxsize_active)
        self.w("no_return_permission").set_active(0)

    def do_maxsize_toggled(self, widget):
        self.w("maxsize_entry").set_sensitive(widget.get_active())
    def do_IMG_toggled(self, widget):
        self.w("IMT_entry").set_sensitive(widget.get_active())

    def do_select(self, sel):
        model, iter = sel.get_selected()
        if not iter:
            self.select_none()
            return
        entry = model.get_value(iter, 1)
        r = entry.idRecord
        assert(r)
        self.currentEntry = entry
        entry.subscribe(self.entry_changed)
        self.select(entry)

    def freeze(self):
        self.frozen += 1
    def thaw(self):
        assert self.frozen > 0
        self.frozen -= 1
        if not self.frozen and self.needUpdate:
            self.needUpdate = False
            self.select(self.currentEntry)

    def entry_changed(self, entry):
        if self.frozen > 0:
            self.needUpdate = True
        else:
            self.select(self.currentEntry)

    def select_none(self):
        if self.currentEntry:
            self.currentEntry.unsubscribe(self.entry_changed)
        self.currentEntry = None

        self.w("ID_label").set_text("")
        self.w("addresses_label").set_text("")
        self.w("name_label").set_text("")
        self.w("petname_entry").set_text("")
        #self.w("IMT_entry")
        #self.w("maxsize_entry")
        #self.w("accept_html")
        self.w("outbound_permission_type").set_text(\
            "Their Restrictions on our Outbound")
        self.w("outbound_IMT").set_text("")
        self.w("outbound_maxsize").set_text("")
        self.w("outbound_html").set_text("")
        self.w("outbound_ticket_urls_label").set_text("")

        disabled = ["petname_entry",
                    "ID_label", "addresses_label", "name_label",
                    "transport_label",
                    "vouch_for", "no_return_permission",
                    "grant_specific",
                    "IMT_toggle", "IMT_entry",
                    "maxsize_toggle", "maxsize_entry",
                    "accept_html", "grant_3rd_party",
                    "addressbook_entry_delete",
                    "addressbook_entry_ping"]
        self.set_sensitive(disabled, 0)

    def select(self, entry):
        r = entry.getProperty
        # names and such
        fields = ["petname_entry",
                  "ID_label", "addresses_label", "name_label",
                  "transport_label",
                  "vouch_for", "no_return_permission",
                  "grant_specific"]
        self.set_sensitive(fields, 1)

        self.w("ID_label").set_text(r("id"))
        self.w("addresses_label").set_text(", ".join(r("addresses")))
        self.w("name_label").set_text(r("name") or "")
        self.w("petname_entry").set_text(r("petname") or "")
        self.w("transport_label").set_text(str(r("transports") or "<None>"))
        self.w("vouch_for").set_active(r("vouchFor"))
        no_grant = not r("grantReturnPermission")
        self.w("no_return_permission").set_active(no_grant)

        # TODO: extensions
        #set_image(self.w("entry_image"),
        #          entry.idRecord.extensions.get("image"))

        # inbound: theirPermission
        inbound = ["IMT_toggle", "maxsize_toggle", "accept_html",
                   "grant_3rd_party"]
        p = entry.getTheirPermission("rfc822")
        if p:
            self.w("grant_specific").set_active(1)
            self.set_sensitive(inbound, 1)
        else:
            # display the genericPermission values read-only
            p = self.owner.getGenericPermission("rfc822")
            self.w("grant_specific").set_active(0)
            self.set_sensitive(inbound, 0)
            self.set_sensitive(["IMT_entry", "maxsize_entry"], 0)
        pr = p.getProperty

        if pr("minIMG") != None:
            self.w("IMT_toggle").set_active(1)
            self.w("IMT_entry").set_value(pr("minIMG"))
            self.w("IMT_entry").set_sensitive(1)
        else:
            self.w("IMT_toggle").set_active(0)
            self.w("IMT_entry").set_sensitive(0)

        if pr("maxsize") != None:
            self.w("maxsize_toggle").set_active(1)
            maxsize = float(pr("maxsize") / 1000)
            self.w("maxsize_entry").set_value(maxsize)
            self.w("maxsize_entry").set_sensitive(1)
        else:
            self.w("maxsize_toggle").set_active(0)
            self.w("maxsize_entry").set_sensitive(0)

        self.w("accept_html").set_active(pr("acceptHTML"))
        self.w("grant_3rd_party").set_active(pr("canGrantPermission"))

        # outbound: ourPermission
        p = entry.getOurPermission("rfc822")
        ptype = entry.getOurPermissionType("rfc822")
        if p:
            self.w("outbound_permission_type").set_text(\
                "Their Restrictions on our Outbound: %s" % ptype)
        else:
            self.w("outbound_permission_type").set_text(\
                "They do not accept inbound messages.")

        if p:
            pr = p.getProperty
            minIMG = pr("minIMG")
            if minIMG == None:
                self.w("outbound_IMT").set_text("unlimited")
            else:
                self.w("outbound_IMT").set_text("%d seconds" % minIMG)
            maxsize = pr("maxsize")
            if maxsize == None:
                self.w("outbound_maxsize").set_text("unlimited")
            else:
                self.w("outbound_maxsize").set_text("%d kb" % \
                                                    (maxsize / 1000))
            if pr("acceptHTML"):
                self.w("outbound_html").set_text("Accepts HTML mail")
            else:
                self.w("outbound_html").set_text("Does NOT accept HTML mail")
            urls = ", ".join(p.listTicketTakerURLs())
            self.w("outbound_ticket_urls_label").set_text(urls)
            # we do not bother to display .canGrantPermission

        self.set_sensitive(["addressbook_entry_cancel",
                            "addressbook_entry_ok"], 0)
        self.set_sensitive(["addressbook_entry_delete",
                            "addressbook_entry_ping"], 1)

    def petname_entry_changed(self, widget):
        if not self.currentEntry:
            return
        entry = self.currentEntry
        new_petname = widget.get_text()
        if not new_petname:
            new_petname = None
        # TODO: insure its uniqueness
        entry.setProperty("petname", new_petname)
        # the setProperty triggers our 'updated' callback. Re-select the
        # entry we just changed. self.updated does scheduleSave.
        self.go_to_entry(entry)

    def entry_dirty(self, *args):
        self.set_sensitive(["addressbook_entry_cancel",
                            "addressbook_entry_ok"], 1)

    def do_cancel(self, widget):
        self.select(self.currentEntry)

    def do_ok(self, widget):
        entry = self.currentEntry
        self.freeze()

        entry.setProperty("vouchFor", self.w("vouch_for").get_active())
        return_permission = not self.w("no_return_permission").get_active()
        entry.setProperty("grantReturnPermission", return_permission)

        if self.w("grant_specific").get_active():
            p = self.gui.agent.newPermission("rfc822")

            # set fields
            if self.w("IMT_toggle").get_active():
                IMG = int(self.w("IMT_entry").get_value())
            else:
                IMG = None
            p.setProperty("minIMG", IMG)
            if self.w("maxsize_toggle").get_active():
                maxsize = int(self.w("maxsize_entry").get_value()) * 1000
                p.setProperty("maxsize", maxsize)
            else:
                p.setProperty("maxsize", None)
            p.setProperty("acceptHTML", self.w("accept_html").get_active())
            p.setProperty("canGrantPermission",
                          self.w("grant_3rd_party").get_active())
            entry.setTheirPermission("rfc822", p)
        else:
            entry.revokeTheirPermission("rfc822")

        self.gui.agent.log("marking entry as modified")
        self.thaw() # this triggers a re-select


    def do_delete(self, widget):
        # confirm
        xml = gtk.glade.XML(gladefile, "confirm_dialog")
        dialog = xml.get_widget("confirm_dialog")
        l = xml.get_widget("confirm_label")
        confirm = ""
        entry = self.currentEntry
        if not entry:
            return
        if entry.theirPermission:
            confirm += "Their specific Permission will be revoked.\n"
        confirm += "\n"
        confirm += "<b>Really delete entry '%s'?</b>\n" % entry.getName()
        l.set_markup(confirm)
        dialog.set_title("Delete Entry %s" % entry.getName())
        dialog.connect("response", self._confirm_response, entry)
    def _confirm_response(self, dialog, responseid, entry):
        if responseid == gtk.RESPONSE_OK:
            self.really_delete(entry)
        dialog.destroy()
    def really_delete(self, entry):
        book = self.owner.getAddressBook()
        book.deleteEntry(entry) # will notifyWatchers when finished

    def do_ping(self, widget):
        entry = self.currentEntry
        if not entry:
            return
        d = self.gui.agent.pingRecipient(entry)
        d.addCallback(self._ping_done, entry)

    def _ping_done(self, response, entry):
        text = "ping response received from %s" % entry.getName()
        d = gtk.MessageDialog(self.gui.xml.get_widget("app"),
                              gtk.DIALOG_DESTROY_WITH_PARENT,
                              gtk.MESSAGE_INFO,
                              gtk.BUTTONS_CLOSE,
                              text)
        d.connect("response", lambda widget, arg: widget.destroy())
        d.show()

    def do_send(self, widget):
        recip = self.currentEntry
        if not recip:
            return
        sender = recip.getOwner()
        sendpane = self.gui.sendpane
        sendpane.do_clear(None)
        sendpane.w("send_frommenu_entry").set_text(sender.getName())
        sendpane.w("to_entry").set_text(recip.getName())
        sendpane.recipientEntry = recip
        self.gui.xml.get_widget("notebook").set_current_page(2)

    def updated(self, addrbook):
        self.populate()
        self.gui.scheduleSave()

class RefsPane(Helper):
    def __init__(self, sendpane):
        xml = gtk.glade.XML(gladefile, "addrefs")
        self.xml = xml
        self.parent = sendpane
        self.window = xml.get_widget("addrefs")
        self.window.connect("destroy", self.do_destroy)
        # activate doesn't work with combo boxes, it seems
        #xml.signal_connect("on_refs_addr_activate", self.do_add)
        xml.signal_connect("do_addrefs_close", self.do_close)
        xml.signal_connect("do_add_ref", self.do_add)
        self.populate()
        self.updateRefs()

    def show(self):
        self.window.deiconify()

    def do_destroy(self, widget):
        self.parent.addrefs_closed()

    def do_add(self, widget):
        name = self.w("refs_adder_entry").get_text().strip()
        ownername = self.parent.w("send_frommenu_entry").get_text()
        owner = self.parent.gui.getOwnerByName(ownername)
        book = owner.getAddressBook()
        r = book.lookupPetName(name)
        if not r:
            r = book.lookupAddress(name)
        if r:
            self.addRef(r)

    def do_close(self, widget):
        self.window.destroy()

    def addRef(self, entry):
        self.parent.refs.append(entry)
        self.updateRefs()

    def updateRefs(self):
        for c in self.w("refs_box").get_children():
            c.destroy()
        for entry in self.parent.refs:
            box = gtk.HBox()
            box.pack_start(gtk.Label(entry.getName()))
            b = gtk.Button("View")
            b.connect("clicked", self.do_view, entry)
            box.pack_end(b, expand=False)
            b = gtk.Button("Remove")
            b.connect("clicked", self.do_remove, entry)
            box.pack_end(b, expand=False)
            box.show_all()
            self.w("refs_box").add(box)

    def do_view(self, widget, entry):
        addrpane = self.parent.gui.addrpane
        addrpane.go_to_entry(entry)
        self.parent.gui.xml.get_widget("notebook").set_current_page(1)

    def do_remove(self, widget, entry):
        self.parent.refs.remove(entry)
        self.updateRefs()

    def populate(self):
        ownername = self.parent.w("send_frommenu_entry").get_text()
        owner = self.parent.gui.getOwnerByName(ownername)
        book = owner.getAddressBook()
        addrs = []
        for e in book.entries:
            addrs.append(e.getName())
        if not addrs:
            addrs = [""]
        default = addrs[0]
        self.w("refs_adder").set_popdown_strings(addrs)
        self.w("refs_adder_entry").set_text(default)


class SendPane(Helper):
    def __init__(self, gui):
        self.gui = gui
        self.recipientEntry = None
        self.refpane = None
        self.refs = []

    def load(self, xml):
        self.xml = xml
        xml.signal_connect("do_send", self.do_send)
        xml.signal_connect("do_addref", self.do_addref)
        xml.signal_connect("do_clear", self.do_clear)
        xml.signal_connect("on_to_entry_changed", self.do_to_changed)
        self.populateOwners()

    def do_to_changed(self, widget):
        self.recipientEntry = None
        self.refs = []
        if self.refpane:
            self.refpane.window.destroy()
        return False

    def populateOwners(self):
        owners = self.gui.getIdentityNames()
        owners = ["-anonymous-"] + owners
        if len(owners) > 1:
            default = owners[1]
        else:
            default = owners[0]
        self.w("send_frommenu").set_popdown_strings(owners)
        self.w("send_frommenu_entry").set_text(default)

    def do_clear(self, button):
        self.recipientEntry = None
        self.w('to_entry').set_text("")
        self.w('subject_entry').set_text("")
        self.w('send_body').get_buffer().set_text("")
        self.refs = []
        if self.refpane:
            self.refpane.window.destroy()

    def do_addref(self, button):
        if not self.refpane:
            self.refpane = RefsPane(self)
        self.refpane.show() # make sure it is on top and not iconified

    def addrefs_closed(self):
        self.refpane = None

    def do_send(self, button):
        print "sending"
        ownername = self.w("send_frommenu_entry").get_text()
        if ownername == "-anonymous-":
            owner = None
        else:
            owner = self.gui.getOwnerByName(ownername)
        if not owner:
            print "Hey, no owner.. anonymous sending not yet implemented"
            return

        m = self.gui.agent.newMessage(mtype="rfc822")
        m.setSender(owner)

        subject = self.w('subject_entry').get_text()
        buf = self.w('send_body').get_buffer()
        start, end = buf.get_bounds()
        body = buf.get_text(start, end)
        fullbody = "From: %s\n" % owner.names + \
                   "To: you\n" + \
                   "Subject: %s\n" % subject + \
                   "\n" + body
        if fullbody[-1] != "\n":
            fullbody += "\n"
        m.addRFC822(fullbody)

        for e in self.refs:
            m.addReference(e)

        m.doDump = True
        if self.recipientEntry:
            m.addRecipient_correspondent(self.recipientEntry)
        else:
            name = self.w('to_entry').get_text().strip()
            # TODO: name might be an address
            ##m.agent.sendMessageToNameOrAddress(name)
            m.addRecipient_petname(name)
        m.send()

FROM = 0
SUBJECT = 1
MESSAGE = 2

class Inbox:
    def __init__(self):
        self.messages = []
    def deliver(self, mtype, msg):
        self.messages.append(msg)
        self.parent.receive(msg)
    def __getstate__(self):
        d = self.__dict__.copy()
        del d['parent']
        return d

class InboxPane(Helper):

    def __init__(self, gui):
        self.gui = gui
        self.currentMessage = None
        self.knownOwners = []
        self.inbox = self.gui.agent.inbox
        self.inbox.parent = self

    def populateOwners(self):
        owners = self.gui.agent.getAllIdentities()
        for o in self.knownOwners[:]:
            if not o in owners:
                o.unsubscribeMailbox(self.inbox.deliver)
                self.knownOwners.remove(o)
        for o in owners:
            if not o in self.knownOwners:
                self.knownOwners.append(o)
                o.subscribeMailbox(self.inbox.deliver)

    def load(self, xml):
        self.xml = xml
        model = gtk.ListStore(str, # from
                              str, # subject
                              gobject.TYPE_PYOBJECT, # msg,
                              )
        self.messages = model

        self.view = view = self.w("inbox_tree")
        view.set_model(model)
        view.append_column(gtk.TreeViewColumn("From",
                                              gtk.CellRendererText(),
                                              text=FROM))
        view.append_column(gtk.TreeViewColumn("Subject",
                                              gtk.CellRendererText(),
                                              text=SUBJECT))
        sel = view.get_selection()
        sel.set_mode(gtk.SELECTION_SINGLE)
        sel.connect("changed", self.do_message)

        xml.signal_connect("on_inbox_delete_clicked", self.do_delete)
        xml.signal_connect("on_inbox_reply_clicked", self.do_reply)
        self.populateOwners()
        self.populate_messages()

    def populate_messages(self):
        # run at startup time to populate the list from the stored inbox
        self.messages.clear()
        for msg in self.inbox.messages:
            self.receive(msg, notify=False)

    def receive(self, msg, notify=True):
        d = msg.getFromName()
        d.addCallback(self._receive2, msg, notify)

    def _receive2(self, fromName, msg, notify):
        self.messages.append((fromName, "<subject>", msg))
        self.gui.scheduleSave()
        if not notify:
            return

        text = ("Message received\n" +
                "\n" +
                "From: %s\n" % fromName +
                "To: %s\n" % msg.getRecipient().getProperty("myName"))
        d = gtk.MessageDialog(self.gui.xml.get_widget("app"),
                              gtk.DIALOG_DESTROY_WITH_PARENT,
                              gtk.MESSAGE_INFO,
                              gtk.BUTTONS_CLOSE,
                              text)
        def close(widget, arg):
            widget.destroy()
            self.gui.sheepbox.plantSheep()
        d.connect("response", close)
        d.show()
        self.gui.sheepbox.plantSheep2()
        
    def do_message(self, sel):
        model, iter = sel.get_selected()
        if not iter:
            self.clear_message()
            return
        msg = model.get_value(iter, MESSAGE)
        self.currentMessage = msg
        self.populate()

    def clear_message(self):
        self.w('from_label').set_text("")
        self.w('to_label').set_text("")
        self.w('from_permission_label').set_text("")
        self.w('inbox_body').get_buffer().set_text("")
        for c in self.w("inbox_buttons").get_children():
            c.destroy()
        for c in self.w('inbox_from_buttons').get_children():
            c.destroy()

    def addButton(self, label, text):
        box = gtk.HBox()
        b = gtk.Button(label)
        box.pack_start(b, expand=False)
        box.pack_end(gtk.Label(text))
        box.show_all()
        self.w("inbox_buttons").add(box)
        return b

    def populate(self):
        msg = self.currentMessage
        d = msg.getFromName()
        d.addCallback(self._populate2, msg)
    def _populate2(self, fromName, msg):
        self.w('from_label').set_text(fromName)
        self.w('to_label').set_text(msg.getRecipient().getProperty("myName"))
        self.w('from_permission_label').set_text(msg.getReason())
        self.w('inbox_body').get_buffer().set_text(msg.getMessage())

        self.populate_buttons()

    def populate_buttons(self):
        msg = self.currentMessage
        buttons = self.w("inbox_buttons")
        # clear the buttonbox
        for c in buttons.get_children():
            c.destroy()
        if msg.canAddSender():
            d = msg.getFromName()
            d.addCallback(self._populate_buttons2, msg)
        if msg.canReply():
            b = self.addButton("Reply",
                               "Reply to this message")
            b.connect("clicked", self.do_reply)
        for r in msg.canAddReferences():
            b = self.addButton("Add Reference",
                               "Add %s to your Addressbook" % r['name'])
            b.connect("clicked", self.do_add_reference, r, msg)

        # add from-buttons
        for c in self.w('inbox_from_buttons').get_children():
            c.destroy()
        if msg.getSender():
            b = gtk.Button("View")
            self.w('inbox_from_buttons').add(b)
            b.connect("clicked", self.do_view_sender, msg)
            b.show()

    def _populate_buttons2(self, fromName, msg):
        b = self.addButton("Add Sender",
                           "Add %s to your Addressbook" % fromName)
        b.connect("clicked", self.do_add_sender, msg)


    def do_add_sender(self, widget, msg):
        d = msg.addSender()
        d.addCallback(self._do_add_sender1)
        self.populate_buttons()

    def _do_add_sender1(self, entry):
        # make the address-pane jump to the entry, but don't actually switch
        # to that pane
        self.gui.addrpane.go_to_entry(entry)
        self.populate_buttons()

    def do_add_reference(self, widget, refnode, msg):
        d = msg.addReference(refnode)
        d.addCallback(self._do_add_sender1)

    def do_view_sender(self, widget, msg):
        addrpane = self.gui.addrpane
        addrpane.go_to_entry(msg.getSender())
        self.gui.xml.get_widget("notebook").set_current_page(1)

    def do_delete(self, widget):
        sel = self.view.get_selection()
        model, iter = sel.get_selected()
        if not iter:
            print "no message selected"
            return
        (fromName, subject, msg) = model[iter]
        del model[iter]
        self.inbox.messages.remove(msg)
        self.view.get_selection().unselect_all()

    def do_reply(self, widget):
        msg = self.currentMessage
        sendpane = self.gui.sendpane
        sender_name = msg.getRecipient().getProperty("myName")
        sendpane.w("send_frommenu_entry").set_text(sender_name)
        d = msg.getFromName()
        d.addCallback(self._reply2, sendpane)
        sendpane.recipientEntry = msg.getSender()
        sendpane.w("subject_entry").set_text("Reply to your message")
        sendpane.w("send_body").get_buffer().set_text("")
        # TODO: also make the TextView active ("give it focus"?)
        self.gui.xml.get_widget("notebook").set_current_page(2)

    def _reply2(self, fromName, sendpane):
        sendpane.w("to_entry").set_text(fromName)

class SheepBox(object):
    def __init__(self, gui):
        self.gui = gui
    def load(self, xml):
        self.area = xml.get_widget("sheeparea")
        self.label = xml.get_widget("sheeplabel")
        self.entry = xml.get_widget("sheepentry")
        self.plantSheep()
        self.showEntry(0)

    def OLDmakeLabel(self, text):
        self.label.set_text(text)
        if self.boxContents == self.label:
            return
        if self.boxContents != None:
            self.boxContents.hide()
            self.box.remove(self.boxContents)
        self.box.add(self.label)
        self.boxContents = self.label
        self.boxContents.show_all()
    def OLDmakeEntry(self, instructions):
        self.vboxInstructions.set_text(instructions)
        if self.boxContents == self.vbox:
            return
        if self.boxContents != None:
            self.boxContents.hide()
            self.box.remove(self.boxContents)
        self.box.add(self.vbox)
        self.boxContents = self.vbox
        self.boxContents.show_all()

    def showEntry(self, show):
        if show:
            self.entry.show()
        else:
            self.entry.hide()

    def plantSheep(self):
        if self.gui.sheep_picture:
            self.area.set_from_file(self.gui.sheep_picture)
            return
        sheep = sibling("sheep.gif")
        if not os.path.exists(sheep):
            sheep = sibling("../../docs/sheep.gif")
        if os.path.exists(sheep):
            self.area.set_from_file(sheep)
    def plantSheep2(self):
        if self.gui.sheep_picture2:
            self.area.set_from_file(self.gui.sheep_picture2)


class SetupPane(Helper):
    def __init__(self, parent):
        self.parent = parent
    def loadXML(self):
        self.xml = xml = gtk.glade.XML(gladefile, "setup")
        self.window = window = xml.get_widget("setup")
        window.connect("destroy", self.do_destroy)
        self.ok_button = xml.get_widget("setup_ok_button")

        fields = ["setup_pop3_port_label", "setup_pop3_port",
                  #"setup_imap_port_label", "setup_imap_port",
                  "setup_smtp_port_label", "setup_smtp_port",
                  ]
        self.proxy_fields = fields

        c = xml.signal_connect
        c("do_dirty", self.do_dirty)

        c("on_setup_address_pb_toggled", self.address_pb_toggled)
        c("on_setup_address_disk_toggled", self.address_disk_toggled)
        
        c("on_setup_nntp_needpass_toggled", self.nntp_needpass_toggled)

        c("on_setup_mixminion_path_activate", self.mm_path_activate)
        c("on_setup_enable_proxy_toggled", self.proxy_toggled)

        c("on_setup_close", self.do_close)
        c("on_setup_ok", self.do_ok)
        
        self.populate()
        self.ok_button.set_sensitive(0)

    def do_close(self, widget):
        self.window.destroy()

    def populate(self):
        agent = self.parent.agent

        lookup = agent.getAddressLookup()
        if lookup.startswith("PB:"):
            self.w("setup_address_pb").set_active(1)
            self.w("setup_address_disk").set_active(0)
            self.w("setup_address_fake").set_active(0)
            # TODO: use an instance instead of these strings
            scheme, host, port = lookup.split(":")
            self.w("setup_address_pb_host").set_text(host)
            self.w("setup_address_pb_port").set_text(port)
        elif lookup.startswith("Disk:"):
            self.w("setup_address_pb").set_active(0)
            self.w("setup_address_disk").set_active(1)
            self.w("setup_address_fake").set_active(0)
        elif lookup.startswith("Fake:"):
            self.w("setup_address_pb").set_active(0)
            self.w("setup_address_disk").set_active(0)
            self.w("setup_address_fake").set_active(1)

        mail = agent.smtpScheme
        assert mail.startswith("Mail:")
        scheme, relayhost, fromaddr = mail.split(":")
        self.w("setup_smtp_host").set_text(relayhost)
        self.w("setup_smtp_from").set_text(fromaddr)

        mme = self.w("setup_mixminion_path")
        self.mm_path, self.mm_version = agent.getMixminionPath()
        if self.mm_path:
            mme.set_text(self.mm_path)
            self.w("setup_mixminion_version").set_text(self.mm_version)
        else:
            if mme.get_text() == "":
                mme.set_text("mixminion")
                self.mm_get_version(quiet=True) # run a version query

        nntp = agent.nntpScheme
        if nntp:
            pass
        self.w("setup_nntp_needpass").set_active(0)

        if agent.pop3_port or agent.imap_port or agent.smtp_port:
            self.w("setup_enable_proxy").set_active(1)
        else:
            self.w("setup_enable_proxy").set_active(0)

        if agent.pop3_port:
            self.w("setup_pop3_port").set_text("%d" % agent.pop3_port)
        else:
            self.w("setup_pop3_port").set_text("")
        if agent.imap_port:
            self.w("setup_imap_port").set_text("%d" % agent.imap_port)
        else:
            self.w("setup_imap_port").set_text("")
        if agent.smtp_port:
            self.w("setup_smtp_port").set_text("%d" % agent.smtp_port)
        else:
            self.w("setup_smtp_port").set_text("")

        # now emit "toggled" signals on all these widgets so they do their
        # update things
        for n in ["setup_address_pb",
                  "setup_address_disk",
                  "setup_address_fake",
                  "setup_enable_proxy",
                  "setup_nntp_needpass"]:
            self.w(n).emit("toggled")
        

    def do_dirty(self, widget):
        self.ok_button.set_sensitive(1)

    def address_pb_toggled(self, widget):
        fields = ["setup_address_pb_host_label", "setup_address_pb_host",
                  "setup_address_pb_port_label", "setup_address_pb_port",
                  ]
        self.set_sensitive(fields, widget.get_active())

    def address_disk_toggled(self, widget):
        fields = ["setup_address_disk_directory_label",
                  "setup_address_disk_directory",
                  ]
        self.set_sensitive(fields, widget.get_active())

    def address_fake_toggled(self, widget):
        fields = ["setup_address_fake_label",
                  ]
        self.set_sensitive(fields, widget.get_active())

    def mm_path_changed(self, widget):
        self.w("setup_mixminion_version").set_sensitive(0)

    def mm_path_activate(self, widget):
        print "activated"
        self.mm_get_version(quiet=False)

    def mm_get_version(self, quiet):
        path = self.w("setup_mixminion_path").get_text()
        d = mm_util.getVersion(path)
        d.addCallback(self._mm_version_done, path, quiet)
        d.addErrback(lambda why: why.printTraceback())

    def _mm_version_done(self, results, path, quiet):
        version, error = results
        if version:
            self.mm_path = path
            self.mm_version = version
            self.w("setup_mixminion_version").set_text(version)
            self.w("setup_mixminion_version").set_sensitive(1)
            self.do_dirty(None)
        else:
            self.w("setup_mixminion_version").set_text("?")
            if quiet:
                return
            text = "Unable to locate mixminion at path\n"
            text += "%s\n" % path
            text += str(error)
            d = gtk.MessageDialog(self.window,
                                  gtk.DIALOG_DESTROY_WITH_PARENT,
                                  gtk.MESSAGE_INFO,
                                  gtk.BUTTONS_CLOSE,
                                  text)
            d.connect("response", lambda widget, arg: widget.destroy())
            d.show()

    def nntp_needpass_toggled(self, widget):
        fields = ["setup_nntp_user_label", "setup_nntp_user",
                  "setup_nntp_pass_label", "setup_nntp_pass",
                  ]
        self.set_sensitive(fields, widget.get_active())

    def proxy_toggled(self, widget):
        self.set_sensitive(self.proxy_fields, widget.get_active())

    def do_ok(self, widget):
        agent = self.parent.agent

        if self.w("setup_address_pb").get_active():
            lookup = "PB:%s:%s" % (self.text("setup_address_pb_host"),
                                   self.text("setup_address_pb_port"))
        elif self.w("setup_address_disk").get_active():
            lookup = "Disk:%s" % self.text("setup_address_disk_directory")
        elif self.w("setup_address_fake").get_active():
            lookup = "Fake:"
        else:
            print "Weird address lookup: nothing selected"
            
        agent.setAddressLookup(lookup)

        relayhost = self.w("setup_smtp_host").get_text()
        smtpfrom = self.w("setup_smtp_from").get_text()
        smtp = "Mail:%s:%s" % (relayhost, smtpfrom)
        agent.setSMTP(smtp)

        # TODO: not quite right, if they changed the path text but didn't
        # hit return, then we haven't done a version-check yet.
        agent.setMixminionPath(self.mm_path, self.mm_version)

        host = self.w("setup_nntp_host").get_text()
        port = self.w("setup_nntp_port").get_text()
        port = int(port)
        if self.w("setup_nntp_needpass").get_active():
            user = self.w("setup_nntp_user").get_text()
            passwd = self.w("setup_nntp_pass").get_text()
            nntp = "%s:%s:%s:%s" % (host, port, user, passwd)
        else:
            nntp = "%s:%s" % (host, port)
        agent.setNNTP(nntp)

        pop3_port, imap_port, smtp_port = None, None, None
        if self.w("setup_enable_proxy").get_active():
            t = self.w("setup_pop3_port").get_text()
            if t:
                pop3_port = int(t)
            t = self.w("setup_imap_port").get_text()
            if t:
                imap_port = int(t)
            t = self.w("setup_smtp_port").get_text()
            if t:
                smtp_port = int(t)
        agent.setProxy(pop3_port, imap_port, smtp_port)

        self.window.destroy()
        self.parent.scheduleSave()

    def show(self):
        self.window.deiconify()
    def do_destroy(self, widget):
        self.parent.closeSetup()

        
class Gui(object):
    def __init__(self, agent):
        self.agent = agent
        self.addrpane = AddressBookPane(self)
        self.mepane = MePane(self)
        self.sendpane = SendPane(self)
        self.inboxpane = InboxPane(self)
        self.sheepbox = SheepBox(self)
        self.logpane = None
        self.saveTimer = None
        self.setup = None
        self.owner_picture = None

    # convenience methods
    def getIdentityNames(self):
        return [i.getProperty("myName") for i in self.agent.getAllIdentities()]
    def getOwnerByName(self, name):
        for i in self.agent.getAllIdentities():
            if i.getProperty("myName") == name:
                return i

    def scheduleSave(self):
        if self.saveTimer:
            self.saveTimer.cancel()
        self.saveTimer = reactor.callLater(5, self.do_save)

    def do_quit(self, widget):
        if self.saveTimer:
            self.saveTimer.cancel()
            self.saveTimer = None
        self.agent.stop()

    def do_setup(self, widget):
        if self.setup:
            self.setup.show()
            return
        self.setup = SetupPane(self)
        self.setup.loadXML()
    def closeSetup(self):
        self.setup = None

    def do_event_log(self, widget):
        if self.logpane:
            self.logpane.show()
        else:
            self.logpane = LogWindow(self)
    def event_log_closed(self):
        print "log closed"
        self.logpane = None

    def do_debug_shell(self, widget):
        print "unimplemented"

    def loadXML(self):
        self.xml = xml = gtk.glade.XML(gladefile, "app")
        xml.signal_connect("do_setup", self.do_setup)
        xml.signal_connect("do_quit", self.do_quit)
        xml.signal_connect("do_event_log", self.do_event_log)
        xml.signal_connect("do_debug_shell", self.do_debug_shell)
        self.addrpane.load(xml)
        self.mepane.load(xml)
        self.sendpane.load(xml)
        self.inboxpane.load(xml)
        self.sheepbox.load(xml)
        self.ticketAgents = [TicketAgent(self)]

    def do_save(self):
        if self.saveTimer:
            self.saveTimer = None
        if self.savefile:
            print "saving to", self.savefile
            tmpfile = self.savefile + ".tmp"
            try:
                pickle.dump(self.agent, open(tmpfile, "w"))
                os.rename(tmpfile, self.savefile)
            except (OSError, IOError), x:
                print "couldn't save:", x

class GuiAgent(agent.Agent):
    gui = None

    def __init__(self, basedir):
        agent.Agent.__init__(self, basedir)
        self.inbox = Inbox()

    def __getstate__(self):
        d = agent.Agent.__getstate__(self)
        del d['gui']
        return d

    def setupGui(self, gui):
        self.gui = gui

    def getTicketAgents(self):
        return self.gui.ticketAgents

rcstring = '''
#pixmap_path "/tmp"
style "pmdefault"
{
 bg[NORMAL] = "%(normal)s"
 bg[ACTIVE] = "%(active)s"
 bg[PRELIGHT] = "%(prelight)s"
 bg[SELECTED] = "%(selected)s"
 bg[INSENSITIVE] = "%(insensitive)s"

 base[NORMAL] = "%(base_normal)s"
 base[ACTIVE] = "%(base_active)s"
 base[PRELIGHT] = "%(base_prelight)s"
 base[SELECTED] = "%(base_selected)s"
 base[INSENSITIVE] = "%(base_insensitive)s"

# bg_pixmap[NORMAL] = "red.png"
}
class "GtkWidget" style "pmdefault"

'''

def set_color(color):
    tint = None
    if color == "red":
        tint = 1.05, .90, .90
    elif color == "blue":
        tint = .95, .95, 1.20
    elif color == "green":
        tint = .95, 1.20, .95
    if tint:
        gray = {"normal": (0xd3, 0xd3, 0xdd),
                "active": (0xc1, 0xc1, 0xcc),
                "prelight": (0xe5, 0xe5, 0xf7),
                "selected": (0x33, 0x66, 0x99),
                "insensitive": (0xd3, 0xd3, 0xdd),
                "base_normal": (0xe5, 0xe5, 0xf7),
                "base_active": (0x80, 0x80, 0x80),
                "base_prelight": (0xe5, 0xe5, 0xe7),
                "base_selected": (0x33, 0x66, 0x99),
                "base_insensitive": (0xe5, 0xe5, 0xf7),
                }
        tinted = gray.copy()
        def mult(a, b):
            r = max(0,min(a[0]*b[0],255))
            g = max(0,min(a[1]*b[1],255))
            b = max(0,min(a[2]*b[2],255))
            return (r,g,b)
        for k in tinted:
            gray = tinted[k]
            tinted[k] = mult(gray, tint)
            tinted[k] = "#%02x%02x%02x" % tinted[k]
        gtk.rc_parse_string(rcstring % tinted)

def run(config):
    basedir = config['basedir']
    testing = config['devel-test']

    set_color(config['color'])
    savefile = os.path.join(basedir, "gui.pickle")
    if os.path.exists(savefile):
        print "loading"
        a = pickle.load(open(savefile, "r"))
        from twisted.persisted import styles
        styles.doUpgrade()
    else:
        print "initializing"
        if not os.path.exists(basedir):
            os.mkdir(basedir)
        a = GuiAgent(basedir)
        a.init()
        if testing:
            a.setAddressLookup("Disk:_DiskNameServer")

    log.addObserver(a.eventLog.addEvent)

    gui = Gui(a)
    gui.sheep_picture = config.get('sheep')
    gui.sheep_picture2 = config.get('sheep2')
    a.setupGui(gui)
    gui.savefile = savefile
    gui.loadXML()
    gui.mepane.populate()
    a.run()
    gui.do_save()
