
"""
Implement a TicketAgent in the gui console.
"""

from twisted.internet import reactor
from twisted.internet.defer import Deferred, DeferredList
from twisted.python import log

from petmail import interfaces, util, crypto
from petmail.server.ticketserver import PBTicketClient

from textwrap import dedent

import gtk.gdk

class RequestClient(PBTicketClient):
    def __init__(self, url, gui, label, entry, area):
        PBTicketClient.__init__(self, url)
        self.gui = gui
        self.label = label
        self.entry = entry
        self.area = area

    def getTicket(self):
        self.label.set_text(dedent("""\
        [%s]
        Retrieving challenge...
        """ % self.url))
        return PBTicketClient.getTicket(self)

    def displayImage(self, image):
        # image is a PNG file
        self.label.set_text(dedent("""\
        [%s]
        Read the letters and
        type them into the
        box below.
        """ % self.url))
        loader = gtk.gdk.gdk_pixbuf_loader_new() # make a GdkPixbufLoader
        loader.write(image)
        loader.close()
        p = loader.get_pixbuf()
        self.area.set_from_pixbuf(p)
        self.entry.set_text("")
        self.entry.show()

    def do_entry(self, widget):
        self.label.set_text(dedent("""\
        [%s]
        Sending response...
        """ % self.url))
        self.sendResponse(widget.get_text())

    def gotTicket(self, ticket):
        self.label.set_text(dedent("""\
        [%s]
        ticket granted""" % self.url))
        self.entry.hide()
        reactor.callLater(5, self.label.set_text,
                          dedent("""\
                          Welcome
                          to
                          PETmail!
                          """))
        reactor.callLater(0.5, self.gui.sheepbox.plantSheep)
        PBTicketClient.gotTicket(self, ticket)
        
class TicketAgent(object):
    __implements__ = interfaces.ITicketAgent
    debug = 0

    def __init__(self, gui):
        self.gui = gui
        self.area = gui.sheepbox.area
        self.label = gui.sheepbox.label
        self.entry = gui.sheepbox.entry
        self.entry.connect("activate", self.do_entry)
        self.activeRequest = None

    def handleRequirements(self, requirements, msg):
        # this must return a list of Deferreds, each of which will be fired
        # with a single Ticket
        newreqs = []
        dl = []
        for r in requirements[:]:
            if r[0] == 'ticket':
                d = self.getTicket(msg, r[1:])
                if d:
                    dl.append(d)
                    requirements.remove(r)
        if dl and self.debug:
            print "handleRequirements returning", dl
        return dl

    def getTicket(self, msg, req):
        """Try to obtain the given ticket. Return a Deferred that the
        message should wait upon."""
            
        urls, = req
        if self.debug:
            print "URLS", urls
        if not msg.headers["rfc822"].has_key("tickets"):
            msg.headers["rfc822"]["tickets"] = []
        url = urls[0]
        request = RequestClient(url, self.gui,
                                self.label, self.entry, self.area)
        self.activeRequest = request
        d = request.getTicket()
        d.addErrback(self.ticketFailed, url)
        return d

    def ticketFailed(self, why, url):
        self.label.set_text(dedent("""\
        [%s]
        ticket DENIED""" % url))
        self.entry.hide()
        print why
        return why

    def do_entry(self, widget):
        self.activeRequest.do_entry(widget)
