#! /usr/bin/python

from twisted.internet import defer
from twisted.application import service
from twisted.python.components import implements

from petmail import interfaces


class Plugin(service.MultiService):
    __implements__ = interfaces.IPlugin, service.MultiService.__implements__

    def outboundMessage(self, msg):
        pass
    def inboundMessage(self, msg):
        pass

class AddressBookEntryPlugin(Plugin):
    __implements__ = interfaces.IEntryPlugin, Plugin.__implements__
    def entry(self):
        return self.parent
    def owner(self):
        return self.entry().parent.parent
    def agent(self):
        return self.owner().parent

    def entryChanged(self, action, entry, lognode):
        pass

class OwnerPlugin(Plugin):
    __implements__ = interfaces.IOwnerPlugin, Plugin.__implements__
    def owner(self):
        return self.parent
    def agent(self):
        return self.owner().parent

    def entryChanged(self, action, entry, lognode):
        pass

class AgentPlugin(Plugin):
    __implements__ = interfaces.IAgentPlugin, Plugin.__implements__
    def agent(self):
        return self.parent
    def ownerChanged(self, action, owner, lognode):
        pass
    def entryChanged(self, action, entry, lognode):
        pass

class PluginHolderMixin:
    acceptablePluginInterface = None

    def addPlugin(self, plugin):
        assert implements(plugin, interfaces.IPlugin)
        if self.acceptablePluginInterface is not None:
            assert implements(plugin, self.acceptablePluginInterface), \
                   "plugin %s is not suitable for addition to %s" % \
                   (plugin, self)
            
        plugin.setServiceParent(self)

    def removePlugin(self, plugin):
        assert plugin in list(self)
        assert implements(plugin, interfaces.IPlugin)
        plugin.disownServiceParent()

    def getPlugins(self):
        for child in list(self):
            if implements(child, interfaces.IPlugin):
                yield child

    def deleteAllPlugins(self):
        for child in self.getPlugins():
            self.removePlugin(child)

    def runPlugins(self, methname, *args):
        l = []
        for child in self.getPlugins():
            meth = getattr(child, methname)
            l.append(defer.maybeDeferred(meth, *args))
        dl = defer.DeferredList(l, fireOnOneErrback=1, consumeErrors=1)
        # if a plugin dies, the DeferredList errback is given a Failure made
        # up of a unholy tuple of the real failure and an index number. We
        # extract the real failure
        dl.addErrback(lambda res: res.value[0])
        return dl

    def runPluginsD(self, res, methname, *args):
        # like runPlugins but suitable for use in a deferred callback
        return self.runPlugins(methname, *args)

