
from twisted.trial import unittest
from twisted.internet import defer
dr = unittest.deferredResult

import petmail.crypto
petmail.crypto.EncryptionAgent = petmail.crypto.FakeEncryptionAgent

from petmail import message, addressbook, owner, agent, permission

class LazyAgent(agent.Agent):
    def scheduleEvent(self, *args):
        return defer.Deferred() # will never fire

class OutboundMessage(unittest.TestCase):
    def build1(self, anon):
        self.agent = a = LazyAgent("test_message.OutboundMessage")
        a.init()
        us = dr(a.createIdentity())
        a.addOwner(us)
        m = message.OutboundMessage(a)
        m.setSenderEntry(us, anon)
        #m.senderEntry = addressbook.AddressBookEntry()
        #m.recipientEntry = recip
        return m

    def testCreate(self):
        m = self.build1(False)
        self.failIf(m.anonymous)

    def testCreateAnon(self):
        m = self.build1(True)
        self.failUnless(m.anonymous)

    def testRFC822(self):
        m = self.build1(False)
        m.addRFC822("some body text\n")

    def testTransport(self):
        raise unittest.SkipTest("need to implement this")
        m = self.build1(False)
"""

    def testTransport1(self):
        # immediate return
        m = message.OutboundMessage(self.agent)
        m.setSenderEntry(FakeOwnerEntry(1, self))
        r = FakeID(2, self)
        r.transport = 3
        m.setRecipientRecord(r)
        self.gettransports = []
        p = processors.Transport()

        self.failIf(m.transport)
        r = p.checkMessage(m)
        dr = unittest.deferredResult(r)
        self.failUnlessEqual(dr, True)
        self.failUnlessEqual(self.gettransports, [(self.agent, m)])
        self.failUnlessEqual(m.transport, 3)

    def testTransport2(self):
        # immediate failure
        m = message.OutboundMessage(self.agent)
        m.setSenderEntry(FakeOwnerEntry(1, self))
        r = FakeID(2, self)
        r.error = True
        m.setRecipientRecord(r)
        self.gettransports = []
        p = processors.Transport()

        self.failIf(m.transport)
        self.failUnlessRaises(Unsendable, p.checkMessage, m)

    def testTransport3(self):
        # deferred return
        m = message.OutboundMessage(self.agent)
        m.setSenderEntry(FakeOwnerEntry(1, self))
        m.setRecipientRecord(FakeID(2, self))
        self.gettransports = []
        p = processors.Transport()

        self.failIf(m.transport)
        r = p.checkMessage(m)
        self.failUnless(isinstance(r, defer.Deferred))
        self.failUnlessEqual(self.gettransports, [(self.agent, m)])
        self.failIf(m.transport)
        
        m.recipientRecord.callback(3)
        self.failUnlessEqual(m.transport, 3)
        dr = unittest.deferredResult(r)
        self.failUnlessEqual(dr, True)

    def testTransport4(self):
        # deferred failure
        m = message.OutboundMessage(self.agent)
        m.setSenderEntry(FakeOwnerEntry(1, self))
        m.setRecipientRecord(FakeID(2, self))
        self.gettransports = []
        p = processors.Transport()

        self.failIf(m.transport)
        r = p.checkMessage(m)
        self.failUnless(isinstance(r, defer.Deferred))
        self.failUnlessEqual(self.gettransports, [(self.agent, m)])
        self.failIf(m.transport)
        
        m.recipientRecord.errback(Unsendable("oops"))
        self.failUnlessEqual(m.transport, None)
        dr = unittest.deferredError(r)
        self.failUnless(dr.check(Unsendable))
"""

class ReceivedMessage(unittest.TestCase):

    def build1(self):
        self.agent = a = LazyAgent("test_message.Message")
        a.init()
        recip = dr(a.createIdentity())
        a.addOwner(recip)
        m = message.ReceivedMessage(a, None)
        m.headers = {}
        m.senderEntry = addressbook.AddressBookEntry()
        m.recipientEntry = recip
        return m

    def setSigned(self, m):
        sender = dr(self.agent.createIdentity())
        sender.setName("Filbert Q. Huntingslope")
        sender.addAddress("pliffle@example.com")
        signed = dr(sender.sign())
        m.headers['senderIDRecord'] = signed
        return sender

    def grantPermission(self, m):
        p = permission.OwnerPermission()
        m.headers['yourPermission'] = p.encode()

    def testCreateIDRecord1(self):
        m = self.build1()
        sender = self.setSigned(m)
        dr(m.createIDRecord())
        r = m.senderIDRecord
        self.failUnless(r)
        self.failUnlessEqual(r.name, sender.names)
        self.failUnlessEqual(r.addresses, sender.addresses)

    def testCreateIDRecord2(self):
        m = self.build1()
        # do not give it a senderIDRecord
        dr(m.createIDRecord())
        self.failIf(m.senderIDRecord)

    def testCanAddSender(self):
        m = self.build1()
        self.failIf(m.canAddSender()) # senderEntry is already there

        m.senderEntry = None
        self.failIf(m.canAddSender()) # no senderIDRecord

        self.setSigned(m)
        self.failUnless(m.canAddSender())

    def testAddSender1(self):
        m = self.build1()
        m.senderEntry = None
        sender = self.setSigned(m)
        assert m.canAddSender()

        e = dr(m.addSender())
        self.failUnless(e)
        self.failUnless(e.addressbook)
        book = self.agent.getAddressBook(m.recipientEntry.ID)
        self.failUnlessEqual(e.addressbook, book)
        self.failUnlessEqual(e.idRecord.ID, sender.ID)
        self.failUnlessEqual(e.idRecord.name, sender.names)
        self.failUnlessEqual(e.idRecord.addresses, sender.addresses)
        self.failUnless(e.theirPermission) # should be granted
        self.failIf(e.ourPermission)

    def testAddSender2(self):
        m = self.build1()
        m.senderEntry = None
        self.setSigned(m)
        self.grantPermission(m)
        assert m.canAddSender()

        e = dr(m.addSender())
        self.failUnless(e)
        self.failUnless(e.addressbook)
        book = self.agent.getAddressBook(m.recipientEntry.ID)
        self.failUnlessEqual(e.addressbook, book)
        self.failUnless(e.theirPermission) # should be granted
        self.failUnless(e.ourPermission)

    def testCanGrantInboundPermission(self):
        m = self.build1()
        self.failUnless(m.canGrantInboundPermission())

        m.senderEntry.theirPermission = True
        self.failIf(m.canGrantInboundPermission()) # already have it

        m.senderEntry = None
        self.failIf(m.canGrantInboundPermission()) # nobody to grant it to

    def testGrantInboundPermission(self):
        m = self.build1()
        self.setSigned(m)
        e = dr(m.addSender())
        e.theirPermission = None

        assert m.canGrantInboundPermission()

        m.grantInboundPermission()
        self.failUnless(e.theirPermission)

    def testGetFromName1(self):
        m = self.build1()
        sender = self.setSigned(m)
        e = dr(m.addSender())

        # this gets it from .senderEntry
        self.failUnlessEqual(dr(m.getFromName()), '"%s"' % sender.names)
        m.senderEntry.setProperty("petname", "bruce")
        self.failUnlessEqual(dr(m.getFromName()), "bruce")

    def testGetFromName2(self):
        m = self.build1()
        m.senderEntry = None
        sender = self.setSigned(m)
        e = dr(m.createIDRecord())
        assert not m.senderEntry

        # this gets it from .senderIDRecord
        self.failUnlessEqual(dr(m.getFromName()), '"%s"' % sender.names)
        m.senderIDRecord.name = None
        self.failUnlessEqual(dr(m.getFromName()), sender.ID)

    def testGetFromName3(self):
        m = self.build1()
        m.senderEntry = None
        m.senderIDRecord = None

        # this message is anonymous
        self.failUnlessEqual(dr(m.getFromName()), "-anonymous-")
        
