
from twisted.trial import unittest
from twisted.spread import pb
from twisted.internet import reactor
dr = unittest.deferredResult

import petmail.crypto
petmail.crypto.EncryptionAgent = petmail.crypto.FakeEncryptionAgent

from petmail import util, ticket, crypto
from petmail.server import ticketserver

class TestClient(ticketserver.PBTicketClient):
    def displayImage(self, image):
        self.sendResponse(image)

class Test(unittest.TestCase):
    def setUp(self):
        self.crypto = crypto.FakeEncryptionAgent()
        privkey = dr(self.crypto.generateKey("ticketserver"), 60)
        server = ticketserver.PBTicketServer(None, privkey, fake=True)
        factory = pb.PBServerFactory(server, unsafeTracebacks=True)
        p = reactor.listenTCP(0, factory)
        portnum = p.getHost().port
        self.url = "PB:localhost:%d" % portnum
        server.url = self.url

    def testGetKey(self):
        d = ticketserver.PBTicketKeyClient(self.url).getKey()
        keydata = unittest.deferredResult(d)
        key = dr(self.crypto.addPubkey(keydata), 60)
        return key

    def testGetTicket(self):
        key = self.testGetKey()
        now = util.now()
        d = TestClient(self.url).getTicket()
        signed = unittest.deferredResult(d)
        recip, sigID, plaintext = dr(self.crypto.decrypt(signed), 60)
        t = util.decode(plaintext)
        self.failUnlessEqual(sigID, key.getID())
        self.failUnlessEqual(t['URL'], self.url)
        self.assertApproximates(now, t['Timestamp'], 5)
        self.failUnless(t['TicketID'])

        taker = ticket.TicketTaker(self.url, None, key)
        self.failUnless(taker.checkTicket(sigID, t))
