
from petmail import util
from twisted.trial import unittest
import time, base64
import StringIO

class Netstring(unittest.TestCase):
    def testNetstring(self):
        tn = util.to_netstring
        tns = util.to_netstrings
        fn = util.from_netstring
        fns = util.from_netstrings

        blob = "abcd"
        self.failUnless(tn(blob) == "4:abcd,")

        blob = "abcdefghijkl"
        self.failUnless(tn(blob) == "12:abcdefghijkl,")

        s = "12:abcdefghijkl,"
        self.failUnless(fn(s) == "abcdefghijkl")

        strings = "3:abc,6:abcdef,0:,2:ab,"
        f = StringIO.StringIO(strings)
        self.failUnless(fn(f) == "abc")
        self.failUnless(fn(f) == "abcdef")
        self.failUnless(fn(f) == "")
        self.failUnless(fn(f) == "ab")
        self.failUnless(fn(f) == None)
        self.failUnless(fn(f) == None)
        
        self.failUnlessRaises(util.InvalidNetstring, fn, ",")
        self.failUnlessRaises(util.InvalidNetstring, fn, "abc")
        self.failUnlessRaises(util.InvalidNetstring, fn, "3:ab,2:de,")
        self.failUnlessRaises(util.InvalidNetstring, fn, "3:abcd,")

        strings = "3:abc,6:abcdef,0:,2:ab,"
        f = StringIO.StringIO(strings)
        g = fns(f)
        self.failUnless(g.next() == "abc")
        self.failUnless(g.next() == "abcdef")
        self.failUnless(g.next() == "")
        self.failUnless(g.next() == "ab")
        self.failUnlessRaises(StopIteration, g.next)
        self.failUnlessRaises(StopIteration, g.next)

        strings = "3:abc,6:abcdef,0:,2:ab,"
        g = fns(strings)
        self.failUnless(g.next() == "abc")
        self.failUnless(g.next() == "abcdef")
        self.failUnless(g.next() == "")
        self.failUnless(g.next() == "ab")
        self.failUnlessRaises(StopIteration, g.next)
        self.failUnlessRaises(StopIteration, g.next)

        self.failUnless(tns(["abc", "defg", ""]) == "3:abc,4:defg,0:,")

class Encode(unittest.TestCase):
    def testList(self):
        d = [1,2,3]
        d2 = util.decode(util.encode(d))
        self.failUnless(d2 == d)
    def testDict(self):
        d = {'Foo': "blah", 'Bar': "bloo boo"}
        d2 = util.decode(util.encode(d))
        self.failUnless(d2 == d)
    def testListOfDict(self):
        d = [{1:2,3:4}, "a", [5,6], [7,[8,9], [], 10]]
        d2 = util.decode(util.encode(d))
        self.failUnless(d2 == d)

class Dict(unittest.TestCase):
    def testEncode1(self):
        d = {'Foo': "blah", 'Bar': "bloo boo"}
        s = util.encodeDict(d)
        self.failUnless(s == "3:Bar,8:bloo boo,3:Foo,4:blah,")
    def testEncode2(self):
        d = {'Num': 123}
        self.failUnlessRaises(ValueError, util.encodeDict, d)
    def testEncode3(self):
        d = {'Oops:': "ok colon"}
        s = util.encodeDict(d)
        self.failUnless(s == "5:Oops:,8:ok colon,")
    def testDecode1(self):
        s = "5:Bluuf,4:bore,3:Oof,3:yar,"
        d = util.decodeDict(StringIO.StringIO(s))
        self.failUnless(d == {'Oof': "yar", 'Bluuf': "bore"})
    def testDecode2(self):
        s = "5:Bluuf,4:bore,3:Oof,3:yar,"
        d = util.decodeDict(s)
        self.failUnless(d == {'Oof': "yar", 'Bluuf': "bore"})

class List(unittest.TestCase):
    def testEncode(self):
        enc = util.encodeList
        self.failUnless(enc([]) == "")
        self.failUnless(enc(["ab"]) == "2:ab,")
        self.failUnless(enc(["ab", "cde"]) == "2:ab,3:cde,")
        self.failUnless(enc(["ab", "", "cde"]) == "2:ab,0:,3:cde,")
        self.failUnlessRaises(ValueError, enc, [12, 34])
    def testDecode(self):
        dec = util.decodeList
        self.failUnless(dec("") == [])
        self.failUnless(dec("2:ab,") == ["ab"])
        self.failUnless(dec("2:ab,3:cde,") == ["ab", "cde"])
        self.failUnless(dec("2:ab,0:,3:cde,") == ["ab", "", "cde"])
        self.failUnless(dec("2:ab,3:cde,0:,") == ["ab", "cde", ""])
        f = StringIO.StringIO("2:ab,3:cde,4:fghi,")
        self.failUnless(dec(f) == ["ab", "cde", "fghi"])
        f = "2:ab,3:cde,4:fghi,"
        self.failUnless(dec(f) == ["ab", "cde", "fghi"])

class Util(unittest.TestCase):
    def tearDown(self):
        util.timeOffset = 0

    def testNow(self):
        start = int(time.time())
        n = util.now()
        stop = int(time.time())
        self.failUnless(start <= n <= stop)
    def testOffset(self):
        util.timeOffset = 15
        start = int(time.time()+15)
        n = util.now()
        stop = int(time.time()+15)
        self.failUnless(start <= n <= stop)

class Armor(unittest.TestCase):
    def testArmor(self):
        msg = "hi\x00\n"
        a = util.armor(msg, "BOOGIE")
        self.failUnless(a.startswith("-----BEGIN PETMAIL BOOGIE-----\n\n"))
        self.failUnless(a.endswith("\n-----END PETMAIL BOOGIE-----\n"))
        start = len("-----BEGIN PETMAIL BOOGIE-----\n\n")
        end = -len("-----END PETMAIL BOOGIE-----\n")
        body = a[start:end]
        m2 = base64.decodestring(body)
        self.failUnlessEqual(msg, m2)

                        
    def testDearmor(self):
        msg = "hi\x00\n" * 90
        a = util.armor(msg, "BOOGIE")
        self.failUnless(util.is_armored(a, "BOOGIE"))
        self.failIf(util.is_armored(a, "NOOGIE"))
        self.failIf(util.is_armored("I am not armor", "BOOGIE"))
        a = "\n\rgarbage\n" + a + "more garbage"
        self.failUnless(util.is_armored(a, "BOOGIE"))
        m2 = util.dearmor(a, "BOOGIE")
        self.failUnlessEqual(msg, m2)
        self.failUnlessRaises(ValueError,
                              util.dearmor, "I am not armor", "BOOGIE")
