#! /usr/bin/python

from twisted.python.components import implements
from twisted.application import service

from petmail import pbutil, interfaces, mm_util, util
from petmail.interfaces import Unsendable

# the recipient adds an Inlet to their agent
# the Inlet produces a transport description record (.encode) (a tuple)
# the transport record is published in the ID Record
# the sender gets the ID Record and extracts the transport record
# the transport record has a prefix and the args
# the prefix is looked up in knownTransports and used to create a Transport

knownTransports = {}
def registerTransport(transporttype, isSecure, factory):
    knownTransports[transporttype,isSecure] = factory

def makeTransport(trecord, message, wantSecure=False):
    if not trecord:
        return Null(None, message)
    ttype = trecord[0]
    factory = knownTransports.get((ttype, wantSecure))
    if not factory and not wantSecure:
        # secure transports are useable too, just slower
        factory = knownTransports.get((ttype, True))
    if not factory:
        raise Unsendable("no transport for type %s" % ttype)
    transport = factory(trecord, message)
    return transport

class BaseInlet(service.MultiService):
    __implements__ = (interfaces.IInlet, service.MultiService.__implements__)

    def setServiceParent(self, parent):
        assert implements(parent, interfaces.IAgent)
        service.MultiService.setServiceParent(self, parent)
    def encode(self):
        raise NotImplementedError("subclass must provide .encoded()")
    def gotMessage(self, encoded, source=(None, None)):
        # source is (identity, surbid)
        self.parent.receiveMessage(encoded, (self, source[0], source[1]))
    def canDecryptType3(self):
        return self.parent.canDecryptType3()
    def decryptType3(self, mmencrypted):
        return self.parent.decryptType3(mmencrypted)

    def getType(self):
        return self.itype


class UsenetDropbox(object):

    """This subclass posts the message to a Usenet group, probably
    alt.anonymous.messages, where the recipient can pick it up at their
    (fairly anonymous) leisure. The sender is not particularly anonymous
    unless they send the message through a remailer network."""

    def __init__(self, trecord, message):
        t, self.newsgroup = trecord
        
    pass
#registerTransport("Usenet", False, UsenetDropbox)


# TODO: the SMTP Tranports have two kinds of configurable state: that which
# comes from the sender's configuration, and that which comes from the
# recipient's TransportRecord.

# recipient defines:
#  "I want messages via SMTP at address bob@foo.com"
#  "I want messages to be posted to alt.anonymous.messages with subject
#   '123'"

# sender defines:
#  "All SMTP messages should be sent with a From: address of alice@foo.com"
#  "All SMTP messages should be sent through smarthost smtp.isp.com"
#  "All SMTP messages should be sent through mixminion"
#  "Usenet posts should be posted to nntp.isp.com"
#  "Usenet posts should be mail-to-newsed through mixminion and m2n.blah.com"

# Likewise, the Inlet contains state that is purely local, which should not
# be published in their TransportRecord. The published data is listed above.

# recipient-local data:
#  "I get SMTP mail by sucking POP from user/pass @pop.isp.com every 5min"
#  " or by IMAP"
#  "I get SMTP by listening on port 25"
#  "I get SMTP by sucking on a maildir at ~/Maildir-petmail every 5 minutes"
#  "I get SMTP by watching ~/Maildir-petmail for changes with FAM or such"
#  "I get SMTP by one of the above, and sometimes doing nymserver-IMAP
#    polling when the nymserver tells me there are a lot of messages queued,
#    and oh by the way please keep the nymserver supplied with SURBs"
#  "I get Usenet posts by polling nntp.isp.com every 30 minutes"
#  "I get SMTP by listening on localhost:9990 or the ~/.petmail-inlet UNIX
#    socket, because my .forward/.qmail says '|nc localhost 9990'"
