#! /usr/bin/python

from twisted.internet import protocol, defer, reactor
from twisted.protocols import imap4

from petmail import interfaces, pbutil
from petmail.transport import base, mail

# IMAP grabber

class IMAPPollProtocol(imap4.IMAP4Client):
    def serverGreeting(self, caps):
        # caps is a dictionary. caps['AUTH'] is a list of login methods
        d = self.login(self.factory.username, self.factory.password)
        d.addCallback(self.loggedIn)
        d.addCallback(self.factory.complete)
        d.addErrback(self.factory.failed)

    def loggedIn(self, *args):
        if self.factory.doDelete:
            d = self.select(self.factory.folder) # read-write
        else:
            d = self.examine(self.factory.folder)
        d.addCallback(self.examine_cb)
        return d

    def examine_cb(self, info):
        if info['EXISTS']:
            d = self.fetchHeaders("1:*")
            d.addCallback(self.gotHeader)
            return d
        else:
            return self.done()

    def gotHeader(self, msgs):
        ids = msgs.keys()
        ids.sort()
        wanted = imap4.MessageSet()
        for msgid in ids:
            headertext = msgs[msgid]['RFC822.HEADER']
            m = email.message_from_string(headertext)
            print "%s: %s" % (msgid, m['subject'])
            if m['subject'] == "PETMAIL MESSAGE":
                wanted.add(msgid)
        print "wanted:", wanted
        if wanted:
            d = self.fetchBody(wanted)
            d.addCallback(self.gotBody, wanted)
            return d
        else:
            return self.done()

    def gotBody(self, msgs, wanted):
        print "gotBody"
        ids = msgs.keys()
        ids.sort()
        for msgid in ids:
            text = msgs[msgid]['RFC822.TEXT']
            self.factory.inlet.gotText(text)
        if self.factory.doDelete:
            print "deleting:", wanted
            # TODO: twisted-1.1.2 will let us remove the str()
            d = self.setFlags(str(wanted), ("\\Deleted",))
            d.addCallback(self.didDelete)
            return d
        else:
            return self.done()

    def didDelete(self, *args):
        print "didDelete", args
        return self.done()

    def done(self):
        d = self.close()
        d.addCallback(self._closed)
        return d

    def _closed(self, results):
        self.transport.loseConnection()

class IMAPPollerFactory(protocol.ClientFactory):
    protocol = IMAPPollProtocol

    def __init__(self, username, password, folder):
        self.username = username
        self.password = password
        self.folder = folder # might be an empty string
        self.deferred = defer.Deferred()
        self.done = False

    def clientConnectionFailed(self, connector, reason):
        self.failed(reason)

    def clientConnectionLost(self, connector, reason):
        self.failed(reason)

    def complete(self, results):
        print "complete"
        self.done = True
        self.deferred.callback(results)

    def failed(self, why):
        if not self.done:
            self.done = True
            print "failure", why
            self.deferred.errback(why)

class IMAPPoller(mail.PollingSMTPInlet):
    doDelete = True
    __implements__ = (mail.PollingSMTPInlet.__implements__,
                      interfaces.IInlet2_IMAP)
    itype = "imap"

    def __init__(self, hostname, portnum, username, password, pollinterval,
                 folder, address):
        mail.PollingSMTPInlet.__init__(self)
        self.hostname = hostname
        self.portnum = portnum
        self.username = username
        self.password = password
        self.pollinterval = pollinterval
        self.address = address
        self.folder = folder # might be an empty string
        self.description = "SMTP:%s" % address

    def encode(self):
        return ("SMTP", self.address)

    def __cmp__(self, other):
        return (cmp(type(self), type(other)) or
                cmp(self.__class__, other.__class__) or
                cmp((self.hostname, self.portnum, self.username,
                     self.password, self.pollinterval, self.folder,
                     self.address),
                    (other.hostname, other.portnum, other.username,
                     other.password, other.pollinterval, other.folder,
                     other.address)))

    def __hash__(self):
        return hash((self.__class__, self.hostname, self.portnum, self.username,
                     self.password, self.pollinterval, self.folder,
                     self.address))

    def check(self):
        print "connecting.."
        f = IMAPPollerFactory(self.username, self.password, self.folder)
        f.doDelete = self.doDelete
        f.inlet = self
        reactor.connectTCP(self.hostname, self.portnum, f)
        return f.deferred

    def getProperty(self, pname):
        if pname in ("hostname", "portnum", "username", "password",
                     "pollinterval", "folder", "address"):
            return getattr(self, pname)
        raise IndexError("%s has no property named %s" % (self, pname))

