#! /usr/bin/python

# flag to help an importer determine if we actually imported the whole thing
# or not. The correct idiom to load this (possibly-failing because of a
# missing 'twibber' module) is:
#
# try:
#    from petmail.transport import jabber
# except ImportError:
#    # happens the first time the interpreter tries to import it
#    jabber = None
# if jabber and not jabber.supported:
#    # happens second and later times
#    jabber = None

supported = False

from twisted.python import log
from twisted.internet import reactor
from twibber import client, elements, xmppstream, xmlstream

from petmail import util, interfaces
from petmail.transport import base, mail

class Client(client.XMPPClient):
    shouldDoPresence = True

    def authSuccess(self):
        try:
            self._authSuccess()
        except:
            log.err()

    def _authSuccess(self):
        # the parent method does .getRoster, which retrieves the roster and
        # manipulates self._roster . self._roster.getPresence(jid) is
        # useful.
        client.XMPPClient.authSuccess(self)
        #self.myJid = self.factory.myJid
        print "we've authd", self.getJID()

        if self.shouldDoPresence:
            # the base class does *not* register our own presence. The
            # server will not send us presence updates (for members of our
            # roster) until we have done so.
            self.sendPresence(priority='-100',
                              # really we don't want IM clients to know about
                              # us at all
                              show='xa',
                              status='petmail agent',
                              )
            # In addition, I don't think it will it send us any messages
            # from other clients until we've announced our presence, or at
            # least until we've retrieved the roster.

            # using Client without sendPresence results in a connection
            # which does receive messages. Using DebugClientReceive without
            # sendPresence does not. ??

        if 0:
            # get a feature list from the server
            d = self.sendDiscoInfoRequest('localhost')
            d.addCallback(self._gotFeatures)


    def _gotFeatures(self, res):
        identities, features, items = res
        features.sort()
        for f in features:
            print " " + f

    def connectionLost(self, why):
        self.factory.inlet.connection = None
        client.XMPPClient.connectionLost(self, why)


    def gotDiscoInfoAnswer(self, res):
        #print "processing gotDiscoInfoAnswer", res
        q = res.getQueryNode()
        #node = q.getAttribute("node") # should match the one queried
        identities = []
        features = []
        items = []
        # q is a minidom.Element
        for f in q.childNodes:
            if f.nodeName == "identity":
                identities.append((f.getAttribute('category'),
                                   f.getAttribute('type'),
                                   f.getAttribute('name')))
            elif f.nodeName == "feature":
                features.append(f.getAttribute('var'))
            elif f.nodeName == "item":
                jid = f.getAttribute('jid')
                node = f.getAttribute('node')
                name = f.getAttribute('name')
                items.append((jid, node, name))
            else:
                print "weird disco item", f
        return (identities, features, items)
    gotDiscoItemsAnswer = gotDiscoInfoAnswer
    
    def show(self, res, data=None):
        print "%s:" % data, res
    gotDiscoInfoError = show
    gotDiscoItemsError = show

    def gotRoster(self, iq):
        # override just so we can see the resulting roster
        client.XMPPClient.gotRoster(self, iq)
        #print self._roster._contacts

    def gotPresence(self, presence):
        print "%s is %s" % (presence.getFrom(),
                            presence.getType() or 'available')
        
        # .gotSubscribe does automatic accept

        # you must respond to a "subscribe" request (with "subscribed") for
        # it to have effect. You do not have to respond to "unsubscribe",
        # but the permission granted by "subscribed" is sticky, so if you
        # don't "unsubscribed" them then you won't be asked for permission
        # next time.

    def gotMessage(self, m):
        # qualify this somehow, to avoid confusion in case IM clients start
        # talking directly to an Agent.
        if m.getType() != "normal":
            print "gotMessage", m
            return
        encoded = m.getBody()
        encoded = encoded.encode() # convert from unicode to string
        # encoded is a base64 encoded encrypted-message, with headers
        if not util.is_armored(encoded, "MESSAGE"):
            print "not a petmail message", m
            return
        encrypted = util.dearmor(encoded, "MESSAGE")
        # TODO: need to add source=(identity, None)
        return self.factory.inlet.gotMessage(encrypted)


    # DiscoMixin:
    #  sending:
    #   self.sendDiscoInfoRequest(to) or self.sendDiscoItemsRequest(to) will
    #   query others. You can override gotDisco(Info|Items)(Answer|Error),
    #   or just attach a callback/errback to the Deferred that the Request()
    #   methods return.
    #
    #  answering:
    #   DiscoMixin handles 'features' and 'identity' for us, using data
    #    registered with .setIdentity(cat, type, name) and
    #    .registerFeature(feature, handler). I don't know what 'handler' is
    #    used for yet.
    #  .gotDiscoItemsRequest() can be overridden to populate a child list

    # beyond here is leftover code from pre-twibber days

    def _connected(self, res=None):
        print "doing disco"
        self.doDisco("bob@localhost/Home").addBoth(self.show)
        if 0:
            e = domish.Element(("jabber:client", "message"), None,
                               {'to': "bob@localhost/Home",
                                'from': self.myJid.full(),
                                'type': 'chat'})
            e.addUniqueId()
            body = domish.Element(("jabber:client", "body"))
            body.addContent("hi bob, buddy")
            e.addChild(body)
            #<x xmlns='jabber:x:event'><offline/></x>
            offline = domish.Element(("jabber:x:event", "x"))
            offline2 = domish.Element(("jabber:client", "offline"))
            offline.addChild(offline2)
            e.addChild(offline)
            self.send(e)
        if 0:
            # jabberd2 doesn't support this
            self.fetchOfflineItems()


    def _doDisco(self, to, node=None, type="info"):
        # returns a Deferred that fires with three arguments: (ids,
        # features, items)
        #  ids is a list of (category, type, name) tuples
        #  features is a list of feature names (namespace strings)
        #  items is a list of (jid, node=None, name=None) tuples

        d = defer.Deferred()
        r = client.IQ(self.xmlstream, "get")
        e = r.addElement(("http://jabber.org/protocol/disco#%s" % type,
                          "query"))
        if node:
            e['node'] = node
        r.addCallback(self._didIQ, d)
        r.send(to)
        d.addCallback(self._didDisco)
        return d

    def _didDisco(self, res):
        q = res.firstChildElement()
        node = q.getAttribute("node") # should match the one queried
        identities = []
        features = []
        items = []
        for f in q.elements():
            if f.name == "identity":
                identities.append((f['category'], f['type'], f['name']))
            elif f.name == "feature":
                features.append(f['var'])
            elif f.name == "item":
                jid = f['jid']
                node = f.getAttribute('node')
                name = f.getAttribute('name')
                items.append((jid, node, name))
            else:
                print "weird disco item", f.toXml()
        return (identities, features, items)

    def _getServerFeatures(self):
        d = self.doDisco(to="localhost")
        d.addCallback(self._gotFeatures)
        return d

    def _fetchOfflineItems(self):
        d = self.doDisco(None,
                         node='http://jabber.org/protocol/offline',
                         type="items")
        d.addCallback(self.show)

class JabberFactory(client.XMPPClientFactory):
    def __init__(self, *args, **kwargs):
        # turn off the silly file_transfer extension, which opens a
        # fixed listening port
        kwargs['extensions'] = []
        client.XMPPClientFactory.__init__(self, *args, **kwargs)

    def buildProtocol(self, addr):
        p = client.XMPPClientFactory.buildProtocol(self, addr)
        self.inlet.connection = p
        return p

class JabberInlet(base.BaseInlet, util.ComparableMixin):
    compare_attrs = ["server", "username", "passwd", "host", "port"]
    connection = None
    itype = "jabber"

    def __init__(self, server, username, passwd,
                 host, port):
        (self.server, self.username, self.passwd, self.host, self.port) = \
                      (server, username, passwd, host, port)
        base.BaseInlet.__init__(self)

    def encode(self):
        return "Jabber", "%s@%s/%s" % (self.username, self.server,
                                       "petmail-agent")

    def getProperty(self, pname):
        if pname in ("server", "username", "passwd", "host", "port"):
            return getattr(self, pname)
        raise IndexError("%s has no property named %s" % (self, pname))

    def startService(self):
        base.BaseInlet.startService(self)
        f = JabberFactory(self.server, self.username, self.passwd,
                          resource="petmail-agent",
                          auth='legacy digest',
                          )
        f.protocol = Client
        f.inlet = self
        self.f = f
        debug = 0
        f.debug = xmppstream.debug = xmlstream.debug = debug
        #f.myJid = "alice@localhost/Jtest"
        reactor.connectTCP(self.server, self.port, f)

    def stopService(self):
        d = base.BaseInlet.stopService(self)
        # shut down existing connection, and don't try to make a new one
        self.f.stopTrying()
        if self.connection:
            self.connection.transport.loseConnection()
        return d

    def __getstate__(self):
        d = base.BaseInlet.__getstate__(self)
        d['connection'] = None
        return d

class JabberTransport:
    __implements__ = interfaces.ITransport,

    def __init__(self, trecord, message):
        t, self.jid = trecord
        self.message = message

    def send(self, encrypted):
        #m = elements.Message(to, ttype="chat", # ??
        #                     ffrom=self.identity.jid,
        #                     )
        #m.setBody(encoded)

        for inlet in self.message.senderEntry.getInlets():
            if isinstance(inlet, JabberInlet):
                break
        else:
            raise Unsendable("need a JabberInlet to use a JabberTransport")
        encapsulated = mail.encapsulateInText(encrypted)
        c = inlet.connection
        c.sendMessage(self.jid, body=encapsulated,
                      ttype='normal',
                      )

    def __repr__(self):
        return "petmail.transport.Jabber<%s>" % self.jid
base.registerTransport("Jabber", False, JabberTransport)


class DebugClientSendToBob(Client):
    def _authSuccess(self):
        client.XMPPClient.authSuccess(self)
        self.sendPresence()
        print "sending message"
        self.sendMessage("bob@localhost/petmail-agent",
                         body="hello world")
        print "sent message"

class DebugClientReceive(Client):
    def _authSuccess(self):
        client.XMPPClient.authSuccess(self)
        self.sendPresence()

    def gotMessage(self, m):
        body = m.getBody()
        print "got message:", body
        print type(body), len(body)
        s = body.encode()
        print type(s), len(s), s


def debug1():
    from twibber import xmppstream, xmlstream
    f = client.XMPPClientFactory("localhost", "alice", "alice",
                                 resource="Jtest",
                                 auth='legacy digest',
                                 )
    f.protocol = Client
    debug = 0
    f.debug = xmppstream.debug = xmlstream.debug = debug
    #f.myJid = "alice@localhost/Jtest"
    reactor.connectTCP("localhost", 5222, f)
    reactor.run()

def debug2():
    i = JabberInlet("localhost", "alice", "alice", "localhost", 5222)
    i.startService()
    reactor.run()

def debug3(name, c):
    from twibber import xmppstream, xmlstream
    f = client.XMPPClientFactory("localhost", name, name,
                                 resource="Jtest",
                                 auth='legacy digest',
                                 extensions=[],
                                 )
    f.protocol = c
    debug = 0
    f.debug = xmppstream.debug = xmlstream.debug = debug
    #f.myJid = "alice@localhost/Jtest"
    reactor.connectTCP("localhost", 5222, f)
    reactor.run()
    
if __name__ == '__main__':
    import sys
    if sys.argv[1] == "1":
        debug3("alice", DebugClientSendToBob)
    else:
        debug3("bob", DebugClientReceive)

supported = True
