#! /usr/bin/python

from twisted.spread import pb
from twisted.python import log

from petmail import interfaces, pbutil
from petmail.transport import base


class PBInlet(base.BaseInlet, pb.Root):
    __implements__ = (base.BaseInlet.__implements__, pb.Root.__implements__,
                      interfaces.IInlet2_PB)
    itype = "pb"

    def __init__(self, hostname, portnum):
        base.BaseInlet.__init__(self)
        self.description = "PB:%s:%d" % (hostname, portnum)
        self.hostname = hostname
        self.portnum = portnum
        server = pbutil.listen(self, portnum)
        server.setServiceParent(self)

    def encode(self):
        return ("PB", self.hostname, self.portnum)

    def __cmp__(self, other):
        return (cmp(type(self), type(other)) or
                cmp(self.__class__, other.__class__) or
                cmp((self.hostname, self.portnum),
                    (other.hostname, other.portnum)))
    def __hash__(self):
        return hash((self.__class__, self.hostname, self.portnum))

    def remote_message(self, encrypted):
        return self.gotMessage(encrypted)

    def getProperty(self, pname):
        if pname in ("hostname", "portnum"):
            return getattr(self, pname)
        raise IndexError("%s has no property named %s" % (self, pname))


class PBTransport:
    __implements__ = interfaces.ITransport,

    def __init__(self, trecord, message):
        t, self.host, self.portnum = trecord
    def send(self, encapsulated):
        d = pbutil.connect(self.host, self.portnum)
        d.addCallback(self._connected, encapsulated)
        d.addErrback(log.err)
        return d

    def _connected(self, ref, encapsulated):
        d = ref.callRemote("message", encapsulated)
        d.addCallback(self._sent, ref)
        d.addErrback(log.err)
        return d
    def _sent(self, result, ref):
        ref.broker.transport.loseConnection()
    def __repr__(self):
        return "petmail.transport.PB<%s:%d>" % (self.host, self.portnum)
base.registerTransport("PB", False, PBTransport)
