
from __future__ import generators

import time, sha, random, base64, re
try:
    from cStringIO import StringIO
except ImportError:
    from StringIO import StringIO

from twisted.spread import banana, jelly

class InvalidNetstring(Exception):
    pass

timeOffset = 0 # for timing hacks during tests
def now():
    return int(time.time()) + timeOffset
def warpForward(secs):
    global timeOffset
    timeOffset += secs

#msgID = 0
msgID = random.randint(0, 10**9)
def generateMessageID():
    global msgID
    msgID += 1
    return str(msgID)

def encode(o):
    sexp = jelly.jelly(o, taster=jelly.globalSecurity)
    return banana.encode(sexp)
def decode(s):
    sexp = banana.decode(s)
    return jelly.unjelly(sexp, taster=jelly.globalSecurity)
    
def encodeDict(d):
    """Encode a dictionary with (binary) string keys and values into a
    single string"""
    keys = d.keys()
    keys.sort()
    return "".join([to_netstring(k) + to_netstring(d[k]) for k in keys])

def decodeDict(f):
    """Reverse the process of encode(). The input can either be a string or
    an open file (something which responds to .read)"""
    g = from_netstrings(f)
    d = {}
    while 1:
        try:
            k = g.next()
        except StopIteration:
            return d
        v = g.next()
        d[k] = v

def encodeList(l):
    return "".join([to_netstring(v) for v in l])
def decodeList(f):
    return [v for v in from_netstrings(f)]
    
def to_netstring(data):
    if not isinstance(data, type("")):
        raise ValueError, "data must be a string: %s %s" % (type(data), data)
    return "%d:%s," % (len(data), data)

def to_netstrings(l):
    return "".join([to_netstring(data) for data in l])

def from_netstring(f):
    g = from_netstrings(f)
    try:
        return g.next()
    except StopIteration:
        return None

def from_netstrings(f):
    """Return a generator which will iterate over all the netstrings in the
    input.

    @param f: the netstrings to process
    @type f: a C{string} or an open file.
    """

    if isinstance(f, type("")):
        f = StringIO(f)
    length = 0
    while 1:
        c = f.read(1)
        if c == '':
            return
        if c in "0123456789":
            length = length * 10 + int(c)
            continue
        if c == ":":
            s = f.read(length)
            if f.read(1) != ",":
                raise InvalidNetstring, "no trailing comma"
            yield s
            length = 0
            continue
        raise InvalidNetstring, "length %d" % length

def sha_hash(s):
    return sha.new(s).hexdigest()

def armor_header(armortype):
    return "-----BEGIN PETMAIL %s-----" % armortype
def armor_footer(armortype):
    return "-----END PETMAIL %s-----" % armortype

def armor(body, armortype):
    header = armor_header(armortype)
    footer = armor_footer(armortype)
    text = base64.encodestring(body)
    armored = "%s\n\n%s%s\n" % (header, text, footer)
    return armored

def is_armored(armored, armortype):
    i1 = armored.find(armor_header(armortype) + "\n")
    i2 = armored.find(armor_footer(armortype) + "\n")
    return (i1 != -1 and i2 != -1 and i1 < i2)

def dearmor(armored, armortype):
    fp = StringIO(armored)
    HEADER = armor_header(armortype)
    FOOTER = armor_footer(armortype)
    mm = "Malformed message "

    while 1:
        line = fp.readline()
        if not line:
            # EOF too early
            raise ValueError(mm + "(no BEGIN banner)")
        if line.strip() == HEADER:
            line = fp.readline()
            if not line:
                raise ValueError(mm + "(missing separator)")
            if line.strip():
                raise ValueError(mm + "(unexpected headers)")
            break

    body = ""
    while 1:
        line = fp.readline()
        if not line:
            # EOF too early
            raise ValueError(mm + "(no END banner)")
        if line.strip() == FOOTER:
            break
        body += line

    return base64.decodestring(body)


class _None:
    pass

class ComparableMixin:
    """Specify a list of attributes that are 'important'. These will be used
    for all comparison operations."""

    compare_attrs = []

    def __hash__(self):
        alist = [self.__class__] + \
                [getattr(self, name, _None) for name in self.compare_attrs]
        return hash(tuple(alist))

    def __cmp__(self, them):
        if cmp(type(self), type(them)):
            return cmp(type(self), type(them))
        if cmp(self.__class__, them.__class__):
            return cmp(self.__class__, them.__class__)
        assert self.compare_attrs == them.compare_attrs
        self_list= [getattr(self, name, _None) for name in self.compare_attrs]
        them_list= [getattr(them, name, _None) for name in self.compare_attrs]
        return cmp(self_list, them_list)
